import dataclasses
import json
from datetime import datetime, date
from enum import Enum
from typing import Any


def json_serializer(obj: Any) -> str:
    """JSON serializer for objects not serializable by default json code"""

    # https://stackoverflow.com/questions/11875770/how-to-overcome-datetime-datetime-not-json-serializable
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    # if isinstance(obj, list):
    #     return f"[{[str(o) for o in obj]}]"
    if hasattr(obj, "__dict__"):
        return json.dumps(obj.__dict__, default=json_serializer)
    return str(obj)


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)  # type: ignore[arg-type]
        if isinstance(o, Enum):
            return o.value
        if isinstance(o, (datetime, date)):
            return o.isoformat().replace("+00:00", ".000Z")
        if hasattr(o, "to_dict"):
            return o.to_dict()
        return super().default(o)



import dataclasses
from typing import Optional, Dict, Any, List

from helix_fhir_client_sdk.filters.sort_field import SortField
from helix_fhir_client_sdk.function_types import RefreshTokenFunction

from helixcore.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)


@dataclasses.dataclass
class FhirReceiverParameters:
    total_partitions: Optional[int]
    batch_size: Optional[int]
    has_token_col: bool
    server_url: Optional[str]
    log_level: Optional[str]
    action: Optional[str]
    action_payload: Optional[Dict[str, Any]]
    additional_parameters: Optional[List[str]]
    filter_by_resource: Optional[str]
    filter_parameter: Optional[str]
    sort_fields: Optional[List[SortField]]
    auth_server_url: Optional[str]
    auth_client_id: Optional[str]
    auth_client_secret: Optional[str]
    auth_login_token: Optional[str]
    auth_scopes: Optional[List[str]]
    auth_well_known_url: Optional[str]
    include_only_properties: Optional[List[str]]
    separate_bundle_resources: bool
    expand_fhir_bundle: bool
    accept_type: Optional[str]
    content_type: Optional[str]
    additional_request_headers: Optional[Dict[str, str]]
    accept_encoding: Optional[str]
    slug_column: Optional[str]
    retry_count: Optional[int]
    exclude_status_codes_from_retry: Optional[List[int]]
    limit: Optional[int]
    auth_access_token: Optional[str]
    resource_type: str
    error_view: Optional[str]
    url_column: Optional[str]
    use_data_streaming: Optional[bool]
    graph_json: Optional[Dict[str, Any]]
    ignore_status_codes: List[int]
    use_id_above_for_paging: Optional[bool]
    pandas_udf_parameters: AsyncPandasUdfParameters
    refresh_token_function: Optional[RefreshTokenFunction] = None
    use_uuid_for_id_above: bool = False

    def set_additional_parameters(
        self, additional_parameters: List[str] | None
    ) -> "FhirReceiverParameters":
        self.additional_parameters = additional_parameters
        return self

    def set_expand_fhir_bundle(
        self, expand_fhir_bundle: bool
    ) -> "FhirReceiverParameters":
        self.expand_fhir_bundle = expand_fhir_bundle
        return self

    def clone(self) -> "FhirReceiverParameters":
        return FhirReceiverParameters(
            total_partitions=self.total_partitions,
            batch_size=self.batch_size,
            has_token_col=self.has_token_col,
            server_url=self.server_url,
            log_level=self.log_level,
            action=self.action,
            action_payload=self.action_payload,
            additional_parameters=self.additional_parameters,
            filter_by_resource=self.filter_by_resource,
            filter_parameter=self.filter_parameter,
            sort_fields=self.sort_fields,
            auth_server_url=self.auth_server_url,
            auth_client_id=self.auth_client_id,
            auth_client_secret=self.auth_client_secret,
            auth_login_token=self.auth_login_token,
            auth_scopes=self.auth_scopes,
            auth_well_known_url=self.auth_well_known_url,
            include_only_properties=self.include_only_properties,
            separate_bundle_resources=self.separate_bundle_resources,
            expand_fhir_bundle=self.expand_fhir_bundle,
            accept_type=self.accept_type,
            content_type=self.content_type,
            additional_request_headers=self.additional_request_headers,
            accept_encoding=self.accept_encoding,
            slug_column=self.slug_column,
            retry_count=self.retry_count,
            exclude_status_codes_from_retry=self.exclude_status_codes_from_retry,
            limit=self.limit,
            auth_access_token=self.auth_access_token,
            resource_type=self.resource_type,
            error_view=self.error_view,
            url_column=self.url_column,
            use_data_streaming=self.use_data_streaming,
            graph_json=self.graph_json,
            ignore_status_codes=self.ignore_status_codes,
            refresh_token_function=self.refresh_token_function,
            use_id_above_for_paging=self.use_id_above_for_paging,
            pandas_udf_parameters=self.pandas_udf_parameters,
            use_uuid_for_id_above=self.use_uuid_for_id_above,
        )



from typing import Optional, cast, Any, Dict

import requests
from requests.auth import HTTPBasicAuth


class TokenHelper:
    @staticmethod
    def get_oauth_token(
        *, client_id: str, client_secret: str, token_url: str, scope: Optional[str]
    ) -> Optional[str]:
        # Prepare the headers and body for the request
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        data = {"grant_type": "client_credentials"}
        if scope:
            data["scope"] = scope

        # Make the POST request to the token endpoint
        response = requests.post(
            token_url,
            headers=headers,
            data=data,
            auth=HTTPBasicAuth(client_id, client_secret),
        )

        # Check if the request was successful
        if response.status_code == 200:
            token_data = response.json()
            return cast(Optional[str], token_data.get("access_token"))
        else:
            raise Exception(
                f"Failed to get token: {response.status_code}, {response.text}"
            )

    @staticmethod
    def get_authorization_header(
        *, client_id: str, client_secret: str, token_url: str, scope: Optional[str]
    ) -> Dict[str, Any]:
        access_token: Optional[str] = TokenHelper.get_oauth_token(
            client_id=client_id,
            client_secret=client_secret,
            token_url=token_url,
            scope=scope,
        )
        assert access_token
        return {"Authorization": f"Bearer {access_token}"}

    @staticmethod
    def get_auth_server_url_from_well_known_url(
        *, well_known_url: str
    ) -> Optional[str]:
        try:
            well_known_response = requests.get(well_known_url)
            # Get token endpoint
            well_known_info = well_known_response.json()
            token_url: Optional[str] = well_known_info.get("token_endpoint")
            return token_url
        except Exception as e:
            return None



from enum import Enum


class LogLevel(Enum):
    ERROR = "ERROR"
    INFO = "INFO"
    DEBUG = "DEBUG"
    TRACE = "TRACE"

    @staticmethod
    def from_str(text: str) -> "LogLevel":
        if text.upper() == "ERROR":
            return LogLevel.ERROR
        if text.upper() == "INFO":
            return LogLevel.INFO
        if text.upper() == "DEBUG":
            return LogLevel.DEBUG
        if text.upper() == "TRACE":
            return LogLevel.TRACE
        raise NotImplementedError(f"{text} is not INFO, DEBUG or TRACE")



import dataclasses
from typing import List, Optional, Any, Dict

from helixcore.utilities.fhir_helpers.fhir_get_response_schema import (
    FhirGetResponseSchema,
)


@dataclasses.dataclass
class FhirGetResponseItem:
    def __init__(self, item: Dict[str, Any]) -> None:
        self.partition_index: int = item.get(FhirGetResponseSchema.partition_index) or 0
        self.sent: int = item.get(FhirGetResponseSchema.sent) or 0
        self.received: int = item.get(FhirGetResponseSchema.received) or 0
        self.responses: List[str] = item.get(FhirGetResponseSchema.responses, [])
        self.first: Optional[str] = item.get(FhirGetResponseSchema.first)
        self.last: Optional[str] = item.get(FhirGetResponseSchema.last)
        self.error_text: Optional[str] = item.get(FhirGetResponseSchema.error_text)
        self.url: Optional[str] = item.get(FhirGetResponseSchema.url)
        self.status_code: Optional[int] = item.get(FhirGetResponseSchema.status_code)
        self.request_id: Optional[str] = item.get(FhirGetResponseSchema.request_id)
        self.access_token: Optional[str] = item.get(FhirGetResponseSchema.access_token)
        self.extra_context_to_return: Optional[str] = item.get(
            FhirGetResponseSchema.extra_context_to_return
        )

    def to_dict(self) -> Dict[str, Any]:
        return self.__dict__

    @classmethod
    def from_dict(cls, item: Dict[str, Any]) -> "FhirGetResponseItem":
        return cls(item)



import uuid
from contextlib import asynccontextmanager, contextmanager
from logging import Logger
from typing import (
    Optional,
    Dict,
    Any,
    AsyncGenerator,
    Generator,
)

from helixcore.logger.yarn_logger import get_logger
from helixcore.utilities.telemetry.telemetry_context import (
    TelemetryContext,
)

from helixcore.utilities.telemetry.console_telemetry_span_wrapper import (
    ConsoleTelemetrySpanWrapper,
)
from helixcore.utilities.telemetry.telemetry import (
    Telemetry,
)
from helixcore.utilities.telemetry.telemetry_parent import (
    TelemetryParent,
)
from helixcore.utilities.telemetry.telemetry_span_wrapper import (
    TelemetrySpanWrapper,
)


class TelemetrySpanCreator:
    def __init__(
        self,
        *,
        telemetry: Optional[Telemetry],
        telemetry_context: TelemetryContext,
        log_level: str = "DEBUG",
    ) -> None:
        """
        Create a telemetry span creator that can create a telemetry span if telemetry is available else return a null context

        :param telemetry: Optional telemetry object
        """

        # Unique instance identifier
        self._instance_id = str(uuid.uuid4())

        assert telemetry is not None
        assert telemetry_context is not None
        self.telemetry = telemetry
        self._current_telemetry_context = telemetry_context

        self._logger: Logger = get_logger(
            __name__,
            level=log_level,
        )
        # get_logger sets the log level to the environment variable LOGLEVEL if it exists
        self._logger.setLevel(log_level)

    def __getstate__(self) -> Dict[str, Any]:
        # Exclude certain properties from being pickled otherwise they cause errors in pickling
        return {k: v for k, v in self.__dict__.items() if k not in ["telemetry"]}

    @asynccontextmanager
    async def create_telemetry_span(
        self,
        *,
        name: str,
        attributes: Optional[Dict[str, Any]],
        telemetry_parent: Optional[TelemetryParent] = None,
    ) -> AsyncGenerator[TelemetrySpanWrapper, None]:
        """
        Create a telemetry span if telemetry is available else return a null context

        :param name: name of the span
        :param attributes:  optional attributes to add to the span
        :param telemetry_parent: telemetry parent
        :return: AsyncContextManager[Any]
        """

        if self.telemetry is not None:
            span: TelemetrySpanWrapper
            async with self.telemetry.trace_async(
                name=name,
                attributes=attributes,
                telemetry_parent=telemetry_parent,
            ) as span:
                yield span
        else:
            yield ConsoleTelemetrySpanWrapper(
                name=name,
                attributes=attributes,
                telemetry_context=TelemetryContext.get_null_context(),
                telemetry_parent=telemetry_parent,
            )

    @contextmanager
    def create_telemetry_span_sync(
        self,
        *,
        name: str,
        attributes: Optional[Dict[str, Any]],
        telemetry_parent: Optional[TelemetryParent] = None,
    ) -> Generator[TelemetrySpanWrapper, None, None]:
        """
        Create a telemetry span if telemetry is available else return a null context

        :param name: name of the span
        :param attributes:  optional attributes to add to the span
        :param telemetry_parent: telemetry parent
        :return: AsyncContextManager[Any]
        """
        if self.telemetry is not None:
            span: TelemetrySpanWrapper
            with self.telemetry.trace(
                name=name,
                attributes=attributes,
                telemetry_parent=telemetry_parent,
            ) as span:
                yield span
        else:
            yield ConsoleTelemetrySpanWrapper(
                name=name,
                attributes=attributes,
                telemetry_context=TelemetryContext.get_null_context(),
                telemetry_parent=telemetry_parent,
            )

    async def flush_async(self) -> None:
        """
        Flush the telemetry

        :return: None
        """
        if self.telemetry:
            await self.telemetry.flush_async()



import dataclasses

from dataclasses_json import DataClassJsonMixin


@dataclasses.dataclass
class TelemetryParent(DataClassJsonMixin):
    name: str
    """ Name of the telemetry parent """

    trace_id: str
    """ Trace ID for the telemetry context """

    span_id: str
    """ Span ID for the telemetry context """



from logging import Logger
from types import TracebackType
from typing import Any, Dict, List, Type, Sequence, Optional, override

from helixcore.utilities.telemetry.telemetry_span_creator import (
    TelemetrySpanCreator,
)
from helixcore.utilities.metrics.base_metrics import (
    BaseMetric,
)
from helixcore.utilities.metrics.writer.base_metrics_writer_async import (
    BaseMetricsWriterAsync,
)
from helixcore.utilities.metrics.writer.base_metrics_writer_parameters import (
    BaseMetricsWriterParameters,
)
from helixcore.utilities.mysql.my_sql_writer.v2.my_sql_writer import (
    MySqlWriter,
)


class MetricsWriter(BaseMetricsWriterAsync):
    def __init__(
        self,
        *,
        parameters: BaseMetricsWriterParameters,
        logger: Optional[Logger],
        telemetry_span_creator: TelemetrySpanCreator,
    ) -> None:
        """
        This class writes metrics to the database

        :param logger: logger to use
        :param parameters: parameters for the metrics writer"""
        super().__init__(
            parameters=parameters,
            logger=logger,
            telemetry_span_creator=telemetry_span_creator,
        )
        self.my_sql_writer: Optional[MySqlWriter] = None

    @override
    async def __aenter__(self) -> "MetricsWriter":
        self.my_sql_writer = MySqlWriter(schema_name=self.schema_name, max_batch_size=0)
        await self.my_sql_writer.open_async()
        return self

    @override
    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        if self.my_sql_writer is not None:
            await self.my_sql_writer.close_async()

    def _get_table_for_metric(self, *, metric: BaseMetric) -> Optional[str]:
        """
        Gets the table for this metrics from the mapping

        :param metric: metric to get the table for
        :return: table name
        """
        if metric.get_name() in self.metric_table_map:
            return self.metric_table_map[metric.get_name()]
        return None

    def _has_table_been_created_for_metric(self, *, metric: BaseMetric) -> bool:
        """
        Checks whether we have already created the table for this metric

        :param metric: metric to check
        :return: True if the table has been created for this metric
        """
        if metric.get_name() in self.tables_created_for_metric:
            return self.tables_created_for_metric[metric.get_name()]
        return False

    @override
    async def create_table_if_not_exists_async(self, *, metric: BaseMetric) -> None:
        """
        Creates the table if it does not exist

        :param metric: metric to create the table for

        :return: None
        """
        async with self.telemetry_span_creator.create_telemetry_span(
            name="create_table_if_not_exists_async",
            attributes={"metric": metric.get_name()},
        ):
            assert metric, "metric should not be None"

            assert (
                self.my_sql_writer
            ), "my_sql_writer should not be None.  Use this class as a context manager"

            table_name: Optional[str] = self._get_table_for_metric(metric=metric)
            if not table_name:
                return

            create_ddl: str = metric.get_create_ddl(
                db_schema_name=self.schema_name, db_table_name=table_name
            )
            assert create_ddl, "create_ddl should not be None"

            if not self.has_database_been_created:
                await self.my_sql_writer.create_database_async(logger=self.logger)
                self.has_database_been_created = True

            await self.my_sql_writer.run_query_async(
                query=create_ddl, logger=self.logger
            )
            self.tables_created_for_metric[metric.get_name()] = True

    @override
    async def write_single_metric_to_table_async(
        self, *, metric: BaseMetric
    ) -> Optional[int]:
        """
        Writes a single metric to the database

        :param metric: metric to write
        :return: number of rows affected
        """
        return await self.write_metrics_to_table_async(metrics=[metric])

    @override
    async def write_metrics_to_table_async(
        self, *, metrics: Sequence[BaseMetric]
    ) -> Optional[int]:
        """
        Writes the data to the table

        :param metrics: list of metrics to write
        :return: number of rows affected
        """
        assert metrics is not None, "metrics should not be None"

        assert (
            self.my_sql_writer
        ), "my_sql_writer should not be None.  Use this class as a context manager"

        if not any(metrics):
            return 0  # nothing to do

        first_metric: BaseMetric = next(iter(metrics))
        assert first_metric, "first_metric should not be None"

        async with self.telemetry_span_creator.create_telemetry_span(
            name="write_metrics_to_table_async",
            attributes={
                "metric": first_metric.get_name(),
                "number_of_metrics": len(metrics),
            },
        ):

            columns: List[str] = first_metric.columns

            assert columns, "columns should not be None"
            table_name: Optional[str] = self._get_table_for_metric(metric=first_metric)
            if not table_name:
                return 0

            assert len(columns) > 0, "columns should not be empty"
            if not self._has_table_been_created_for_metric(metric=first_metric):
                await self.create_table_if_not_exists_async(metric=first_metric)

            data: List[Dict[str, Any]] = [metric.to_dict() for metric in metrics]

            rows_affected: Optional[int] = (
                await self.my_sql_writer.write_to_table_async(
                    table_name=table_name,
                    columns=columns,
                    data=data,
                    logger=self.logger,
                    create_table_ddl=first_metric.get_create_ddl(
                        db_schema_name=self.schema_name, db_table_name=table_name
                    ),
                )
            )

            return rows_affected

    @override
    async def read_metrics_from_table_async(
        self, metric: BaseMetric
    ) -> List[Dict[str, Any]]:
        """
        Reads the data from the table

        :param metric:
        :return: data read from the table
        """
        async with self.telemetry_span_creator.create_telemetry_span(
            name="read_metrics_from_table_async",
            attributes={"metric": metric.get_name()},
        ):
            assert (
                self.my_sql_writer
            ), "my_sql_writer should not be None.  Use this class as a context manager"

            table_name = self._get_table_for_metric(metric=metric)
            if not table_name:
                return []

            return await self.my_sql_writer.read_from_table_async(
                table_name=table_name, columns=metric.columns
            )



from os import environ
from sys import stdout
from logging import Logger, getLogger, StreamHandler, Formatter, INFO
from typing import Union, TextIO, Optional


def get_logger(
    name: str,
    level: Union[int, str] = INFO,
    formatter: Optional[Formatter] = None,
    stream: Optional[TextIO] = stdout,
) -> Logger:
    logger: Logger = getLogger(name)
    level = level or environ.get("LOGLEVEL") or INFO
    logger.setLevel(level)

    if logger.handlers:
        pass
    else:
        stream_handler: StreamHandler[TextIO] = StreamHandler(stream)
        stream_handler.setLevel(level=level)
        if formatter is None:
            # noinspection SpellCheckingInspection
            # https://docs.python.org/3.1/library/logging.html#formatter-objects
            formatter = Formatter(
                "%(asctime)s %(levelname)s %(module)s.%(funcName)s[%(lineno)d]: %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)
    return logger



import os
import urllib.parse

from helixcore.utilities.aws.config import get_ssm_config


class DocumentDbServerUrl:
    CERT_PATH = "/helix.pipelines/certs/rds-combined-ca-bundle.pem"

    def get_server_url(self) -> str:
        env_name = os.environ.get("ENV", "local").lower()
        server_url: str = ""

        if env_name in ["local"]:
            # document db is not available locally so just use the local mongo server
            server_url = "mongodb://mongo:27017/"
        elif env_name in ["dev"]:
            server_url = self._get_dev_connection_string()
        elif env_name in ["client-sandbox"]:
            server_url = self._get_client_sandbox_connection_string()
        elif env_name in ["staging"]:
            server_url = self._get_staging_connection_string()
        elif env_name in ["prod"]:
            server_url = self._get_prod_connection_string()
        else:
            raise Exception(f"Unexpected environment: {env_name}")

        return server_url

    def _get_dev_connection_string(self) -> str:
        path: str = os.environ.get(
            "MONGO_PASSWORD_SSM_PATH", "/dev-ue1/integrationhub-service/"
        )
        ssm_config = get_ssm_config(path=path)
        password = ssm_config[f"{path}mongo_password"]
        escape_password = urllib.parse.quote_plus(password)
        mongo_username = os.environ.get(
            "MONGO_USERNAME", "dev-svc-integration-hub-service"
        )
        mongo_host = os.environ.get(
            "MONGO_HOST",
            "mongodb+srv://cl-dev-services-pl-0.vpsmx.mongodb.net/?authSource=admin&retryWrites=false",
        )
        return self._get_connection_string(mongo_username, escape_password, mongo_host)

    def _get_client_sandbox_connection_string(self) -> str:
        path: str = os.environ.get(
            "MONGO_PASSWORD_SSM_PATH", "/client-sandbox-ue1/integrationhub-service/"
        )
        ssm_config = get_ssm_config(path=path)
        password = ssm_config[f"{path}mongo_password"]
        escape_password = urllib.parse.quote_plus(password)
        mongo_username = os.environ.get(
            "MONGO_USERNAME", "cs-svc-integration-hub-service"
        )
        mongo_host = os.environ.get(
            "MONGO_HOST",
            "mongodb+srv://cl-clients-sandbox-services-pl-1.bonmy.mongodb.net/?authSource=admin",
        )
        return self._get_connection_string(mongo_username, escape_password, mongo_host)

    def _get_staging_connection_string(self) -> str:
        path: str = os.environ.get(
            "MONGO_PASSWORD_SSM_PATH", "/staging/integrationhub-service/"
        )
        ssm_config = get_ssm_config(path=path)
        password = ssm_config[f"{path}mongo_password"]
        escape_password = urllib.parse.quote_plus(password)
        mongo_username = os.environ.get(
            "MONGO_USERNAME", "staging-svc-integration-hub-service"
        )
        mongo_host = os.environ.get(
            "MONGO_HOST",
            "mongodb+srv://cl-staging-services-pl-3.arsv0.mongodb.net/?authSource=admin&retryWrites=false",
        )
        return self._get_connection_string(mongo_username, escape_password, mongo_host)

    def _get_prod_connection_string(self) -> str:
        path: str = os.environ.get(
            "MONGO_PASSWORD_SSM_PATH", "/prod-ue1/integrationhub-service/"
        )
        ssm_config = get_ssm_config(path=path)
        password = ssm_config[f"{path}mongo_password"]
        escape_password = urllib.parse.quote_plus(password)
        mongo_username = os.environ.get("MONGO_USERNAME", "prod-svc-integration-hub")
        mongo_host = os.environ.get(
            "MONGO_HOST",
            "mongodb+srv://cl-prod-services-pl-0.21mit.mongodb.net/?authSource=admin&retryWrites=false",
        )
        return self._get_connection_string(mongo_username, escape_password, mongo_host)

    def _get_connection_string(self, user: str, password: str, host: str) -> str:
        split_host = host.split("//")
        print_safe_connection_string = f"mongodb+srv://{user}:******@{split_host[1]}"
        print(f"connecthub conn string: {print_safe_connection_string}")
        connection_string = f"mongodb+srv://{user}:{password}@{split_host[1]}"
        return connection_string



import csv
import os


# Define constants
AWS_REGION = os.environ.get("AWS_REGION", "us-east-1")

DYNAMODB_DDL_TABLE_NAME = os.environ.get(
    "DYNAMODB_DDL_TABLE_NAME", "ingestion-table-ddl"
)
DYNAMODB_COLUMN_MAP_TABLE_NAME = os.environ.get(
    "DYNAMODB_COLUMN_MAP_TABLE_NAME", "ingestion-table-conversions"
)
DYNAMODB_LOGGING_TABLE_NAME = os.environ.get(
    "DYNAMODB_LOGGING_TABLE_NAME", "ingestion-verification"
)
DYNAMODB_FILE_METADATA_TABLE_NAME = os.environ.get(
    "DYNAMODB_FILE_METADATA_TABLE_NAME", "ingestion-client-file-metadata"
)
DYNAMODB_CONFIG_TABLE_NAME = os.environ.get(
    "DYNAMODB_CONFIG_TABLE_NAME", "ingestion-config"
)

SNS_ARN_DATABELT_DEVS = "arn:aws:sns:us-east-1:400686897767:Bwell_databelt_dev"
SNS_ARN_INGESTION_VERIFICATION = (
    "arn:aws:sns:us-east-1:400686897767:Bwell_Data_ingestion_file_verification"
)


# Global, single installation of the bwell dialect
class DatabeltCSVDialect(csv.Dialect):
    quoting = csv.QUOTE_ALL
    delimiter = "|"

    quotechar = '"'
    doublequote = True
    skipinitialspace = False
    lineterminator = "\n"


csv.register_dialect("databelt", DatabeltCSVDialect)



from typing import (
    Optional,
    Dict,
    Any,
)

from helixcore.utilities.telemetry.telemetry_context import (
    TelemetryContext,
)

from helixcore.utilities.telemetry.telemetry_parent import (
    TelemetryParent,
)
from helixcore.utilities.telemetry.telemetry_span_wrapper import (
    TelemetrySpanWrapper,
)


class ConsoleTelemetrySpanWrapper(TelemetrySpanWrapper):
    @property
    def span_id(self) -> Optional[str]:
        return (
            self._telemetry_parent.span_id
            if self._telemetry_parent is not None
            else None
        )

    @property
    def trace_id(self) -> Optional[str]:
        return (
            self._telemetry_parent.trace_id
            if self._telemetry_parent is not None
            else None
        )

    def __init__(
        self,
        *,
        name: str,
        attributes: Optional[Dict[str, Any]],
        telemetry_context: Optional[TelemetryContext],
        telemetry_parent: Optional[TelemetryParent],
    ) -> None:
        super().__init__(
            name=name,
            attributes=attributes,
            telemetry_context=telemetry_context,
            telemetry_parent=telemetry_parent,
        )



import os
from logging import Logger
from typing import Dict, Any, Optional, List, cast

import pymysql
import sqlparse
from pymysql import OperationalError, MySQLError, Connection
from pymysql.cursors import Cursor
from pymysql.constants import CLIENT

from helixcore.utilities.mysql.pydatabelt.mysql import (
    get_mysql_config,
    construct_mysql_connection_string,
)


class MySqlWriter:
    def __init__(self, *, schema_name: str) -> None:
        """
        This class writes to MySql databases using the pymysql package


        :param schema_name: name of the schema to write to
        """
        assert schema_name, "schema_name should not be None"
        assert isinstance(schema_name, str), "schema_name should be a string"

        env: str = os.getenv("ENV", "local")

        self.db_config: Dict[str, Any] = get_mysql_config(
            "warehouse", env=env, default_schema=schema_name
        )
        self.jdbc_url: str = construct_mysql_connection_string(
            params={"rewriteBatchedStatements": "true"},
            protocol="jdbc:mysql",
            **self.db_config,
        )
        self.username = self.db_config["username"]
        self.password = self.db_config["password"]
        self.host = self.db_config["host"]
        self.port = int(self.db_config["port"])
        self.schema_name: str = schema_name

    def drop_database(self, *, logger: Optional[Logger] = None) -> None:
        """
        Deletes the database

        :param logger: logger to use
        :return: None
        """
        self.run_query_with_schema(
            query=f"DROP DATABASE IF EXISTS {self.schema_name}",
            logger=logger,
            schema_name=None,
        )

    def create_database(self, *, logger: Optional[Logger] = None) -> None:
        """
        Creates the database

        :param logger: logger to use
        :return: None
        """
        self.run_query_with_schema(
            query=f"CREATE DATABASE IF NOT EXISTS {self.schema_name}",
            logger=logger,
            schema_name=None,
        )

    def recreate_database(self, *, logger: Optional[Logger] = None) -> None:
        """
        Recreates the database

        :param logger: logger to use
        :return: None
        """
        self.drop_database(logger=logger)
        self.create_database(logger=logger)

    def run_query(
        self, *, query: str, logger: Optional[Logger] = None
    ) -> Optional[int]:
        return self.run_query_with_schema(
            query=query, logger=logger, schema_name=self.schema_name
        )

    def run_query_with_schema(
        self,
        *,
        query: str,
        logger: Optional[Logger] = None,
        schema_name: Optional[str],
        params: Optional[Dict[str, Any]] = None,
    ) -> Optional[int]:
        """
        Runs the query


        :param query: query to run
        :param logger: logger to use
        :param schema_name: name of the schema to use
        :param params: parameters to pass to the query
        :return: number of rows affected
        """
        # noinspection PyUnresolvedReferences
        connection: pymysql.Connection[Cursor] = (
            pymysql.connect(
                user=self.username,
                password=self.password,
                host=self.host,
                port=self.port,
                db=schema_name,
                client_flag=CLIENT.MULTI_STATEMENTS,
            )
            if schema_name
            else pymysql.connect(
                user=self.username,
                password=self.password,
                host=self.host,
                port=self.port,
                client_flag=CLIENT.MULTI_STATEMENTS,
            )
        )
        try:
            with connection.cursor() as cursor:
                sanitized_queries = self.sanitize_query(query)
                rows_affected: Optional[int] = 0
                for sanitized_query in sanitized_queries:
                    if logger:
                        logger.debug(f"Executing MySQL query: {sanitized_query}")
                    rows_affected = cursor.execute(sanitized_query, params)
                    connection.commit()
                return rows_affected

        except OperationalError as e:
            if logger:
                logger.error(f"Failed to run query {query}")
            raise e

        except MySQLError as e:
            if logger:
                logger.error("MySQL Error: ", e)
            connection.rollback()
            raise e

        finally:
            connection.close()

    def write_to_table(
        self,
        *,
        table_name: str,
        columns: List[str],
        data: List[Dict[str, Any]],
        logger: Optional[Logger] = None,
    ) -> Optional[int]:
        """
        Writes the data to the table

        :param table_name: name of the table to write to
        :param columns: list of columns to write
        :param data: list of dictionaries to write
        :param logger: logger to use
        :return: number of rows affected
        """
        # create the connection

        assert table_name, "table_name should not be None"
        assert columns, "columns should not be None"
        assert data, "data should not be None"
        assert isinstance(columns, list), "columns should be a list"
        assert isinstance(data, list), "data should be a list"
        if len(data) > 0:
            assert all(
                isinstance(d, dict) for d in data
            ), "data should be a list of dictionaries instead it is a list of " + str(
                [type(d) for d in data]
            )

        # noinspection PyUnresolvedReferences
        connection: Connection[Cursor] = pymysql.connect(
            user=self.username,
            password=self.password,
            host=self.host,
            port=self.port,
            db=self.schema_name,
        )
        # Generate the SQL query
        query = f"INSERT INTO {table_name} ({', '.join(columns)}) VALUES ({', '.join(['%s'] * len(columns))})"
        try:
            # Convert list of dictionaries to list of tuples, ensuring the order of values matches the order of columns
            data_tuples = [tuple(d[col] for col in columns) for d in data]

            with connection.cursor() as cursor:
                if logger:
                    logger.debug(f"Executing MySQL query: {query}")
                rows_affected: Optional[int] = 0
                for sanitized_query in self.sanitize_query(query):
                    # Execute the query
                    rows_affected = cursor.executemany(sanitized_query, data_tuples)
                    connection.commit()
                return rows_affected

        except OperationalError as e:
            if logger:
                logger.error(f"Failed to run query {query}")
            raise e

        except MySQLError as e:
            if logger:
                logger.error("MySQL Error: ", e)
            connection.rollback()
            raise e

        finally:
            connection.close()

    def read_from_table(
        self, *, table_name: str, columns: List[str]
    ) -> List[Dict[str, Any]]:
        # create the connection
        # noinspection PyUnresolvedReferences
        connection: Connection[Cursor] = pymysql.connect(
            user=self.username,
            password=self.password,
            host=self.host,
            port=self.port,
            db=self.schema_name,
        )
        cursor: pymysql.cursors.DictCursor
        with connection.cursor(pymysql.cursors.DictCursor) as cursor:
            query = f"SELECT * FROM {table_name}"
            for sanitized_query in self.sanitize_query(query):
                cursor.execute(sanitized_query)
            # dictionary cursor returns list of dictionaries
            rows: List[Dict[str, Any]] = cast(List[Dict[str, Any]], cursor.fetchall())
            return rows

    @staticmethod
    def sanitize_query(query: str) -> List[str]:
        """
        Sanitizes the query by formatting it with proper indentation and whitespace
        and using sqlparse to parse the query


        :param query: SQL
        :return: list of sanitized queries found in query string
        """
        sanitized_queries: List[str] = []
        # Parse the query
        for parsed_query in sqlparse.parse(query):
            # Format the query with proper indentation and whitespace
            formatted_query = sqlparse.format(
                str(parsed_query), reindent=True, keyword_case="upper"
            )

            # Strip any leading or trailing whitespace
            sanitized_query = formatted_query.strip()

            sanitized_queries.append(sanitized_query)

        return sanitized_queries

    def close(self) -> None:
        pass



import asyncio
import threading
import time
from typing import AsyncGenerator, List, TypeVar, Optional, Coroutine, Any

from concurrent.futures import ThreadPoolExecutor


T = TypeVar("T")


class AsyncHelper:
    @staticmethod
    async def collect_items(generator: AsyncGenerator[T, None]) -> List[T]:
        """
        Collects items from an async generator and returns them as a list

        :param generator: AsyncGenerator
        :return: List[T]
        """
        items = []
        async for item in generator:
            items.append(item)
        return items

    @staticmethod
    async def collect_async_data(
        *, async_gen: AsyncGenerator[T, None], chunk_size: int
    ) -> AsyncGenerator[List[T], None]:
        """
        Collects data from an async generator in chunks of size `chunk_size`

        :param async_gen: AsyncGenerator
        :param chunk_size: int
        :return: AsyncGenerator
        """
        chunk1 = []
        async for item in async_gen:
            chunk1.append(item)
            if len(chunk1) >= chunk_size:
                yield chunk1
                chunk1 = []
        # if there are any chunks left yield them
        if chunk1:
            yield chunk1

    @staticmethod
    def run(fn: Coroutine[Any, Any, T], timeout: Optional[float] = None) -> T:
        """
        Runs an async function but returns the result synchronously
        Similar to asyncio.run() but does not create a new event loop if one already exists

        :param fn: Coroutine
        :param timeout: Optional timeout in seconds to wait for loop to finish running
        :return: T
        """
        try:
            # print(f"Getting running loop")
            loop = asyncio.get_running_loop()
            # print(f"Got running loop")
        except RuntimeError:
            # print(f"Creating new event loop")
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        try:
            if loop.is_running() and timeout is not None:
                # print(f"Loop is running so waiting for it to end")
                start_time = time.time()
                while loop.is_running():
                    current_time = time.time()
                    elapsed_time = current_time - start_time

                    if elapsed_time > timeout:
                        # print(f"Timeout {timeout} reached. Exiting loop.")
                        break

                    # print(f"Waiting for loop to end: {elapsed_time}")
                    time.sleep(1)
            # print(f"Running loop")
            result = loop.run_until_complete(fn)
            # print(f"Ran loop")
        except RuntimeError as e:
            if "This event loop is already running" in str(e):
                try:
                    return AsyncHelper.run_in_thread_pool_and_wait(coro=fn)
                except RuntimeError as e2:
                    raise RuntimeError(
                        f"While calling {fn.__name__} there is already an event loop running."
                        "\nThis usually happens because you are calling this function"
                        " from an asynchronous context so you can't just wrap it in AsyncRunner.run()."
                        f"\nEither use `await {fn.__name__}` or"
                        " use nest_asyncio (https://github.com/erdewit/nest_asyncio)."
                        f"\nException: {e}"
                    )
            else:
                raise e
        return result

    @staticmethod
    def run_in_new_thread_and_wait(coro: Coroutine[Any, Any, T]) -> T:
        """
        Runs the coroutine in a new thread and waits for it to finish

        :param coro: Coroutine
        :return: T
        """
        result: Optional[T] = None
        exception: Optional[Exception] = None

        def target() -> None:
            nonlocal result, exception
            try:
                result = asyncio.run(coro)
            except Exception as e:
                exception = e

        thread = threading.Thread(target=target)
        thread.start()
        thread.join()

        if exception:
            raise exception

        # Allow returning None without checking since T may be an Optional type already
        return result  # type: ignore[return-value]

    @staticmethod
    def run_in_thread_pool_and_wait(coro: Coroutine[Any, Any, T]) -> T:
        """
        Runs the coroutine in a thread pool and waits for it to finish

        :param coro: Coroutine
        :return: T
        """
        result: Optional[T] = None
        exception: Optional[Exception] = None

        def target() -> None:
            nonlocal result, exception
            try:
                result = asyncio.run(coro)
            except Exception as e:
                exception = e

        with ThreadPoolExecutor() as executor:
            future = executor.submit(target)
            future.result()  # This will block until the thread completes

        if exception:
            raise exception

        # Allow returning None without checking since T may be an Optional type already
        return result  # type: ignore[return-value]



class FhirMergeResponseSchema:
    """
    This class provides names for columns in FhirMergeResponse

    Should match to https://github.com/icanbwell/helix.fhir.client.sdk/blob/main/helix_fhir_client_sdk/responses/fhir_merge_response.py
    """

    request_id = "request_id"
    url = "url"
    responses = "responses"
    error = "error"
    access_token = "access_token"
    status = "status"
    data = "data"



import typing
from typing import Any, TypeAlias

"""
This is the type alias for the FHIR resource type when it is has been cleaned and is valid R4
"""
FhirResourceType: TypeAlias = typing.OrderedDict[str, Any]

"""
This is the type alias for the FHIR resource type when it is has been received from the FHIR server but 
may not conform to R4 properly yet
"""
FhirReceivedResourceType: TypeAlias = typing.OrderedDict[str, Any]

"""
This is the type alias for the Human API resource type when it is received from the Human API server
"""
HumanApiResourceType: TypeAlias = typing.OrderedDict[str, Any]

# noinspection SpellCheckingInspection
"""
This is the type alias for the FHIR resource type when it is received as DSTU2
"""
# noinspection SpellCheckingInspection
FhirDstu2ResourceType: TypeAlias = typing.OrderedDict[str, Any]

# noinspection SpellCheckingInspection
"""
This is the type alias for the FHIR resource type when it is received as STU3
"""
# noinspection SpellCheckingInspection
FhirStu3ResourceType: TypeAlias = typing.OrderedDict[str, Any]



import boto3
from typing import Dict, Any


def get_ssm_config(
    path: str = "/prod/databelt/",
    region: str = "us-east-1",
    truncate_keys: bool = False,
) -> Dict[str, Any]:
    ssm = boto3.client("ssm", region_name=region)
    param_list = []
    next_token = None
    done = False
    path_len = len(path)
    while not done:
        args = {
            "Path": path,
            "Recursive": True,
            "WithDecryption": True,
        }
        if next_token:
            args["NextToken"] = next_token
        param_response = ssm.get_parameters_by_path(**args)
        param_list.extend(param_response.get("Parameters", []))
        next_token = param_response.get("NextToken", None)
        if not next_token:
            done = True

    params = {}
    for param in param_list:
        if truncate_keys:
            key = param["Name"][path_len:]
        else:
            key = param["Name"]
        params[key] = param["Value"]

    return params


def put_ssm_config(
    config_arguments: Dict[str, Any],
    region: str = "us-east-1",
) -> Dict[str, Any]:
    """
    Update ssm config in AWS

    :param config_arguments: Configurations sent to ssm client
    :param region: default region in aws
    :return refer https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ssm.html#SSM.Client.put_parameter
    doc for return type
    """
    ssm = boto3.client("ssm", region_name=region)
    response: Dict[str, Any] = ssm.put_parameter(**config_arguments)
    return response



from typing import Dict, List
from unittest.mock import AsyncMock, patch

import pytest
from aiohttp import ClientSession, ClientResponse

from helixcore.utilities.api_helper.v2.http_request import (
    HelixHttpRequest,
    RequestType,
    SingleJsonResult,
    ListJsonResult,
    SingleTextResult,
)


@pytest.mark.asyncio
async def test_get_result() -> None:
    url = "http://example.com"
    request = HelixHttpRequest(url=url, request_type=RequestType.GET)

    mock_response = AsyncMock(ClientResponse)
    mock_response.status = 200
    mock_response.json.return_value = {"key": "value"}

    # Create an AsyncMock for the get method that returns the mock_response
    mock_get = AsyncMock(return_value=mock_response)

    with patch.object(ClientSession, "get", mock_get):
        result: SingleJsonResult = await request.get_result_async()
        assert result.status == 200
        assert result.result == {"key": "value"}


@pytest.mark.asyncio
async def test_get_results() -> None:
    url = "http://example.com"
    request = HelixHttpRequest(url=url, request_type=RequestType.GET)

    mock_response = AsyncMock(ClientResponse)
    mock_response.status = 200
    mock_response.json.return_value = [{"key": "value"}]

    # Create an AsyncMock for the get method that returns the mock_response
    mock_get = AsyncMock(return_value=mock_response)

    with patch.object(ClientSession, "get", mock_get):
        result: ListJsonResult = await request.get_results_async()
        assert result.status == 200
        assert result.result == [{"key": "value"}]


@pytest.mark.asyncio
async def test_get_text() -> None:
    url = "http://example.com"
    request = HelixHttpRequest(url=url, request_type=RequestType.GET)

    mock_response = AsyncMock(ClientResponse)
    mock_response.status = 200
    mock_response.text.return_value = "response_text"

    # Create an AsyncMock for the get method that returns the mock_response
    mock_get = AsyncMock(return_value=mock_response)

    with patch.object(ClientSession, "get", mock_get):
        result: SingleTextResult = await request.get_text_async()
        assert result.status == 200
        assert result.result == "response_text"


@pytest.mark.asyncio
async def test_get_response() -> None:
    url = "http://example.com"
    request = HelixHttpRequest(url=url, request_type=RequestType.GET)

    mock_response = AsyncMock(ClientResponse)
    mock_response.status = 200
    mock_response.text.return_value = "response_text"

    # Create an AsyncMock for the get method that returns the mock_response
    mock_get = AsyncMock(return_value=mock_response)

    with patch.object(ClientSession, "get", mock_get):
        response: ClientResponse = await request.get_response_async()
        assert response.status == 200


def test_get_querystring() -> None:
    url = "http://example.com?key=value"
    request = HelixHttpRequest(url=url, request_type=RequestType.GET)
    querystring: Dict[str, List[str]] = request.get_querystring()
    assert querystring == {"key": ["value"]}


def test_to_string() -> None:
    url = "http://example.com"
    request = HelixHttpRequest(url=url, request_type=RequestType.GET)
    assert request.to_string() == "RequestType.GET url=http://example.com"



import pytest
from typing import Dict, Any, List
from helixcore.utilities.fhir_helpers.fhir_merge_response_item import (
    FhirMergeResponseItem,
)
from helix_fhir_client_sdk.responses.fhir_delete_response import FhirDeleteResponse
from helix_fhir_client_sdk.responses.fhir_merge_response import FhirMergeResponse
from helix_fhir_client_sdk.responses.fhir_update_response import FhirUpdateResponse


@pytest.fixture
def sample_merge_response() -> FhirMergeResponse:
    return FhirMergeResponse(
        request_id=None,
        url="Patient/123",
        status=200,
        error=None,
        access_token=None,
        json_data='{"resourceType": "Patient", "id": "123"}',
        responses=[
            {
                "created": True,
                "updated": False,
                "deleted": False,
                "id": "123",
                "uuid": "uuid-123",
                "resourceType": "Patient",
                "sourceAssigningAuthority": "Authority",
                "resource_version": "1",
                "message": "Success",
                "issue": None,
                "error": None,
                "token": "token-123",
                "resource_json": '{"resourceType": "Patient", "id": "123"}',
            }
        ],
    )


@pytest.fixture
def sample_update_response() -> FhirUpdateResponse:
    return FhirUpdateResponse(
        request_id=None,
        responses="",
        access_token=None,
        error=None,
        status=200,
        url="Patient/123",
        resource_type="Patient",
    )


@pytest.fixture
def sample_delete_response() -> FhirDeleteResponse:
    return FhirDeleteResponse(
        request_id=None,
        responses="",
        access_token=None,
        error=None,
        status=200,
        url="Patient/123",
        resource_type="Patient",
    )


def test_fhir_merge_response_item_from_dict() -> None:
    item_dict: Dict[str, Any] = {
        "created": True,
        "updated": False,
        "deleted": False,
        "id": "123",
        "uuid": "uuid-123",
        "resourceType": "Patient",
        "sourceAssigningAuthority": "Authority",
        "resource_version": "1",
        "message": "Success",
        "issue": None,
        "error": None,
        "token": "token-123",
        "resource_json": '{"resourceType": "Patient", "id": "123"}',
    }
    item = FhirMergeResponseItem.from_dict(item_dict)
    assert item.created is True
    assert item.updated is False
    assert item.deleted is False
    assert item.id == "123"
    assert item.uuid == "uuid-123"
    assert item.resourceType == "Patient"
    assert item.sourceAssigningAuthority == "Authority"
    assert item.resource_version == "1"
    assert item.message == "Success"
    assert item.issue is None
    assert item.error is None
    assert item.token == "token-123"
    assert item.resource_json == '{"resourceType": "Patient", "id": "123"}'


def test_fhir_merge_response_item_to_dict() -> None:
    item = FhirMergeResponseItem(
        created=True,
        updated=False,
        deleted=False,
        id_="123",
        uuid="uuid-123",
        resource_type="Patient",
        source_assigning_authority="Authority",
        resource_version="1",
        message="Success",
        issue=None,
        error=None,
        token="token-123",
        resource_json='{"resourceType": "Patient", "id": "123"}',
    )
    item_dict = item.to_dict()
    assert item_dict["created"] is True
    assert item_dict["updated"] is False
    assert item_dict["deleted"] is False
    assert item_dict["id"] == "123"
    assert item_dict["uuid"] == "uuid-123"
    assert item_dict["resourceType"] == "Patient"
    assert item_dict["sourceAssigningAuthority"] == "Authority"
    assert item_dict["resource_version"] == "1"
    assert item_dict["message"] == "Success"
    assert item_dict["issue"] is None
    assert item_dict["error"] is None
    assert item_dict["token"] == "token-123"
    assert item_dict["resource_json"] == '{"resourceType": "Patient", "id": "123"}'


def test_fhir_merge_response_item_from_merge_response(
    sample_merge_response: FhirMergeResponse,
) -> None:
    items = FhirMergeResponseItem.from_merge_response(
        merge_response=sample_merge_response
    )
    assert len(items) == 1
    item = items[0]
    assert item.created is True
    assert item.updated is False
    assert item.deleted is False
    assert item.id == "123"
    assert item.uuid == "uuid-123"
    assert item.resourceType == "Patient"
    assert item.sourceAssigningAuthority == "Authority"
    assert item.resource_version == "1"
    assert item.message == "Success"
    assert item.issue is None
    assert item.error is None
    assert item.token == "token-123"
    assert item.resource_json == '{"resourceType": "Patient", "id": "123"}'


def test_fhir_merge_response_item_from_update_response(
    sample_update_response: FhirUpdateResponse,
) -> None:
    item = FhirMergeResponseItem.from_update_response(
        update_response=sample_update_response
    )
    assert item.error is None
    assert item.status == 200
    assert item.resourceType == "Patient"


def test_fhir_merge_response_item_from_delete_response(
    sample_delete_response: FhirDeleteResponse,
) -> None:
    item = FhirMergeResponseItem.from_delete_response(
        delete_response=sample_delete_response
    )
    assert item.error is None
    assert item.status == 200
    assert item.resourceType == "Patient"


def test_fhir_merge_response_item_from_responses(
    sample_merge_response: FhirMergeResponse,
    sample_update_response: FhirUpdateResponse,
    sample_delete_response: FhirDeleteResponse,
) -> None:
    responses: List[FhirMergeResponse | FhirUpdateResponse | FhirDeleteResponse] = [
        sample_merge_response,
        sample_update_response,
        sample_delete_response,
    ]
    items = FhirMergeResponseItem.from_responses(responses=responses)
    assert len(items) == 3
    assert items[0].resourceType == "Patient"
    assert items[1].resourceType == "Patient"
    assert items[2].resourceType == "Patient"



from logging import Logger
from typing import Optional, List

from helix_fhir_client_sdk.fhir_client import FhirClient


def get_fhir_client(
    logger: Logger,
    server_url: str,
    auth_server_url: Optional[str] = None,
    auth_client_id: Optional[str] = None,
    auth_client_secret: Optional[str] = None,
    auth_login_token: Optional[str] = None,
    auth_access_token: Optional[str] = None,
    auth_scopes: Optional[List[str]] = None,
    log_level: Optional[str] = None,
    auth_well_known_url: Optional[str] = None,
) -> FhirClient:
    assert server_url

    fhir_client: FhirClient = FhirClient()
    fhir_client = fhir_client.url(server_url)

    if log_level:
        fhir_client = fhir_client.log_level(log_level)

    if auth_server_url:
        fhir_client = fhir_client.auth_server_url(auth_server_url)
    if auth_well_known_url:
        fhir_client = fhir_client.auth_wellknown_url(auth_well_known_url)
    if auth_client_id and auth_client_secret:
        fhir_client = fhir_client.client_credentials(auth_client_id, auth_client_secret)
    if auth_login_token:
        fhir_client = fhir_client.login_token(auth_login_token)
    if auth_access_token:
        fhir_client = fhir_client.set_access_token(auth_access_token)
    if auth_scopes:
        fhir_client = fhir_client.auth_scopes(auth_scopes)

    if logger:
        # noinspection PyTypeChecker
        fhir_client = fhir_client.logger(logger)  # type: ignore

    return fhir_client



import pytest
from _pytest.monkeypatch import MonkeyPatch
from botocore.client import BaseClient

from helixcore.utilities.document_db_connection.v1.document_db_connection import (
    DocumentDbServerUrl,
)


@pytest.mark.parametrize(
    "env_name,url_part",
    [
        ("local", "mongodb://mongo:27017/"),
        ("dev", "fake_dev_password"),
        ("prod", "fake_prod_password"),
    ],
)
def test_can_get_server_url_for_environment(
    monkeypatch: MonkeyPatch, ssm_mock: BaseClient, env_name: str, url_part: str
) -> None:

    ssm_mock.put_parameter(  # type: ignore
        Name="/prod-ue1/integrationhub-service/mongo_password",
        Value="fake_prod_password",
        Type="SecureString",
    )

    ssm_mock.put_parameter(  # type: ignore
        Name="/dev-ue1/integrationhub-service/mongo_password",
        Value="fake_dev_password",
        Type="SecureString",
    )
    monkeypatch.setenv("ENV", env_name)
    server_url: str = DocumentDbServerUrl().get_server_url()
    print(server_url)
    assert server_url
    assert url_part in server_url



import dataclasses
from typing import List


class DataFrameType:
    pass


class DataFrameStringType(DataFrameType):
    pass


class DataFrameIntegerType(DataFrameType):
    pass


class DataFrameTimestampType(DataFrameType):
    pass


class DataFrameBooleanType(DataFrameType):
    pass


class DataFrameFloatType(DataFrameType):
    pass


@dataclasses.dataclass
class DataFrameStructField:
    name: str
    data_type: DataFrameType
    nullable: bool = True


@dataclasses.dataclass
class DataFrameStructType:
    def __init__(self, fields: List[DataFrameStructField] = []):
        self.list: List[DataFrameStructField] = fields

    @property
    def fields(self) -> List[DataFrameStructField]:
        return self.list


class DataFrameArrayType(DataFrameType):
    def __init__(self, struct_type: DataFrameStructType) -> None:
        self.struct_type = struct_type



import json
from datetime import datetime
from json import JSONDecodeError
from logging import Logger
from typing import Any, Dict, List, Optional, Union, cast
from typing import (
    Iterable,
    AsyncGenerator,
)


from furl import furl
from helix_fhir_client_sdk.exceptions.fhir_sender_exception import FhirSenderException
from helix_fhir_client_sdk.fhir_client import FhirClient
from helix_fhir_client_sdk.function_types import HandleStreamingChunkFunction
from helix_fhir_client_sdk.responses.fhir_get_response import FhirGetResponse

from helixcore.logger.yarn_logger import get_logger
from helixcore.structures.fhir_receiver.v2.fhir_receiver_parameters import (
    FhirReceiverParameters,
)
from helixcore.structures.fhir_receiver.v2.structures.get_batch_error import (
    GetBatchError,
)
from helixcore.structures.fhir_receiver.v2.structures.get_batch_result import (
    GetBatchResult,
)
from helixcore.utilities.fhir_helpers.fhir_get_response_item import (
    FhirGetResponseItem,
)
from helixcore.utilities.fhir_helpers.fhir_parser_exception import (
    FhirParserException,
)
from helixcore.utilities.fhir_helpers.fhir_receiver_exception import (
    FhirReceiverException,
)
from helixcore.utilities.fhir_helpers.get_fhir_client import (
    get_fhir_client,
)


class FhirReceiverProcessor:
    """
    This class processes the fhir receiver requests.  It has no dependencies on spark and can be used in any async
    environment.

    """

    @staticmethod
    # function that is called for each partition
    async def send_partition_request_to_server_async(
        *,
        partition_index: int,
        rows: Iterable[Dict[str, Any]],
        parameters: FhirReceiverParameters,
    ) -> AsyncGenerator[Dict[str, Any], None]:
        """
        This function processes a partition calling fhir server for each row in the partition

        This has to be a static function to avoid creating a closure around a class
        https://spark.apache.org/docs/latest/rdd-programming-guide.html#passing-functions-to-spark


        :param partition_index: partition index
        :param rows: rows to process
        :param parameters: FhirReceiverParameters
        :return: rows
        """

        json_data_list: List[Dict[str, Any]] = list(rows)
        assert isinstance(json_data_list, list)
        if len(json_data_list) > 0:
            assert isinstance(json_data_list[0], dict)
        assert parameters
        assert isinstance(parameters, FhirReceiverParameters)
        assert parameters.server_url
        assert isinstance(
            partition_index, int
        ), f"partition_index should be an int but is {type(partition_index)}"

        resource_id_with_token_list: List[Dict[str, Optional[str]]] = [
            (
                {
                    "resource_id": r["id"],
                    "access_token": r["token"],
                    parameters.url_column: r[parameters.url_column],
                    parameters.slug_column: r[parameters.slug_column],
                    "resourceType": r["resourceType"],
                }
                if parameters.has_token_col
                and parameters.url_column
                and parameters.slug_column
                else (
                    {
                        "resource_id": r["id"],
                        "access_token": r["token"],
                    }
                    if parameters.has_token_col
                    else {
                        "resource_id": r["id"],
                        "access_token": parameters.auth_access_token,
                    }
                )
            )
            for r in rows
        ]
        async for result in FhirReceiverProcessor.process_with_token_async(
            partition_index=partition_index,
            resource_id_with_token_list=resource_id_with_token_list,
            parameters=parameters,
        ):
            yield result

    @staticmethod
    async def process_with_token_async(
        *,
        partition_index: int,
        resource_id_with_token_list: List[Dict[str, Optional[str]]],
        parameters: FhirReceiverParameters,
    ) -> AsyncGenerator[Dict[str, Any], None]:
        try:
            first_id: Optional[str] = resource_id_with_token_list[0]["resource_id"]
        except IndexError:
            first_id = None

        try:
            last_id: Optional[str] = resource_id_with_token_list[-1]["resource_id"]
        except IndexError:
            last_id = None

        sent: int = len(resource_id_with_token_list)

        if sent == 0:
            yield FhirGetResponseItem(
                dict(
                    partition_index=partition_index,
                    sent=0,
                    received=0,
                    responses=[],
                    first=None,
                    last=None,
                    error_text=None,
                    url=parameters.server_url,
                    status_code=200,
                    request_id=None,
                    access_token=None,
                    extra_context_to_return=None,
                )
            ).to_dict()
            return

        # if batch and not has_token then send all ids at once as long as the access token is the same
        if (
            parameters.batch_size
            and parameters.batch_size > 1
            and not parameters.has_token_col
        ):
            async for r in FhirReceiverProcessor.process_batch_async(
                partition_index=partition_index,
                first_id=first_id,
                last_id=last_id,
                resource_id_with_token_list=resource_id_with_token_list,
                parameters=parameters,
            ):
                yield r
        else:  # otherwise send one by one
            async for r in FhirReceiverProcessor.process_one_by_one_async(
                partition_index=partition_index,
                first_id=first_id,
                last_id=last_id,
                resource_id_with_token_list=resource_id_with_token_list,
                parameters=parameters,
            ):
                yield r

    @staticmethod
    async def process_one_by_one_async(
        *,
        partition_index: int,
        first_id: Optional[str],
        last_id: Optional[str],
        resource_id_with_token_list: List[Dict[str, Optional[str]]],
        parameters: FhirReceiverParameters,
    ) -> AsyncGenerator[Dict[str, Any], None]:
        for resource1 in resource_id_with_token_list:
            async for r in FhirReceiverProcessor.process_single_row_async(
                partition_index=partition_index,
                first_id=first_id,
                last_id=last_id,
                parameters=parameters,
                resource1=resource1,
            ):
                yield r

    @staticmethod
    async def process_single_row_async(
        *,
        partition_index: int,
        first_id: Optional[str],
        last_id: Optional[str],
        resource1: Dict[str, Optional[str]],
        parameters: FhirReceiverParameters,
    ) -> AsyncGenerator[Dict[str, Any], None]:
        id_ = resource1["resource_id"]
        access_token = resource1["access_token"]
        url_ = resource1.get(parameters.url_column) if parameters.url_column else None
        service_slug = (
            resource1.get(parameters.slug_column) if parameters.slug_column else None
        )
        # resource_type = resource1.get("resourceType")
        request_id: Optional[str] = None
        extra_context_to_return: Optional[Dict[str, Any]] = None
        responses_for_logging_only: List[str] = []
        try:
            response: FhirGetResponse
            async for response in FhirReceiverProcessor.send_simple_fhir_request_async(
                id_=id_,
                server_url_=url_ or parameters.server_url,
                service_slug=service_slug,
                server_url=parameters.server_url,
                parameters=parameters,
            ):
                resources: List[str] = []
                errors: List[GetBatchError] = []
                try:
                    batch_result: GetBatchResult = (
                        FhirReceiverProcessor.read_resources_and_errors_from_response(
                            response=response
                        )
                    )
                    responses_for_logging_only += batch_result.resources
                    resources = batch_result.resources
                    errors = batch_result.errors
                except JSONDecodeError as e2:
                    if parameters.error_view:
                        response.error = f"{(response.error or '')}: {str(e2)}"
                    else:
                        raise FhirParserException(
                            url=response.url,
                            message="Parsing result as json failed",
                            json_data=response.responses,
                            response_status_code=response.status,
                            request_id=response.request_id,
                        ) from e2

                error_text = response.error or (
                    "\n".join([e.error_text for e in errors]) if errors else None
                )
                status_code = response.status
                request_url = response.url
                request_id = response.request_id
                extra_context_to_return = response.extra_context_to_return
                access_token = response.access_token
                yield FhirGetResponseItem(
                    dict(
                        partition_index=partition_index,
                        sent=1,
                        received=len(resources),
                        responses=resources,
                        first=first_id,
                        last=last_id,
                        error_text=error_text,
                        url=request_url,
                        status_code=status_code,
                        request_id=request_id,
                        extra_context_to_return=extra_context_to_return,
                        access_token=access_token,
                    )
                ).to_dict()
        except FhirSenderException as e1:
            error_text = str(e1)
            status_code = e1.response_status_code or 0
            request_url = e1.url
            yield FhirGetResponseItem(
                dict(
                    partition_index=partition_index,
                    sent=1,
                    received=len(responses_for_logging_only),
                    responses=responses_for_logging_only,
                    first=first_id,
                    last=last_id,
                    error_text=error_text,
                    url=request_url,
                    status_code=status_code,
                    request_id=request_id,
                    extra_context_to_return=extra_context_to_return,
                    access_token=access_token,
                )
            ).to_dict()

    @staticmethod
    async def process_batch_async(
        *,
        partition_index: int,
        first_id: Optional[str],
        last_id: Optional[str],
        resource_id_with_token_list: List[Dict[str, Optional[str]]],
        parameters: FhirReceiverParameters,
    ) -> AsyncGenerator[Dict[str, Any], None]:
        response: FhirGetResponse
        async for response in FhirReceiverProcessor.send_simple_fhir_request_async(
            id_=[cast(str, r["resource_id"]) for r in resource_id_with_token_list],
            server_url=parameters.server_url,
            server_url_=parameters.server_url,
            parameters=parameters,
        ):
            responses_from_fhir: List[str] = []
            errors: List[GetBatchError] = []
            try:
                batch_result: GetBatchResult = (
                    FhirReceiverProcessor.read_resources_and_errors_from_response(
                        response=response
                    )
                )
                responses_from_fhir = batch_result.resources
                errors = batch_result.errors
            except JSONDecodeError as e1:
                if parameters.error_view:
                    response.error = f"{(response.error or '')}: {str(e1)}"
                else:
                    raise FhirParserException(
                        url=response.url,
                        message="Parsing result as json failed",
                        json_data=response.responses,
                        response_status_code=response.status,
                        request_id=response.request_id,
                    ) from e1

            error_text = response.error or (
                "\n".join([e.error_text for e in errors]) if errors else None
            )
            status_code = response.status
            request_id: Optional[str] = response.request_id
            is_valid_response: bool = True if len(responses_from_fhir) > 0 else False
            yield FhirGetResponseItem(
                dict(
                    partition_index=partition_index,
                    sent=1,
                    received=len(responses_from_fhir) if is_valid_response else 0,
                    responses=responses_from_fhir if is_valid_response else [],
                    first=first_id,
                    last=last_id,
                    error_text=error_text,
                    url=response.url,
                    status_code=status_code,
                    request_id=request_id,
                    access_token=None,
                    extra_context_to_return=None,
                )
            ).to_dict()

    @staticmethod
    async def send_simple_fhir_request_async(
        *,
        id_: Optional[Union[str, List[str]]],
        server_url: Optional[str],
        server_url_: Optional[str],
        service_slug: Optional[str] = None,
        parameters: FhirReceiverParameters,
    ) -> AsyncGenerator[FhirGetResponse, None]:
        url = server_url_ or server_url
        assert url
        logger: Logger = get_logger(
            __name__,
            level=(
                parameters.log_level if parameters and parameters.log_level else "INFO"
            ),
        )
        async for r in FhirReceiverProcessor.send_fhir_request_async(
            logger=logger,
            resource_id=id_,
            server_url=url,
            extra_context_to_return=(
                {parameters.slug_column: service_slug}
                if parameters.slug_column and service_slug
                else None
            ),
            parameters=parameters,
        ):
            yield r

    @staticmethod
    async def send_fhir_request_async(
        *,
        logger: Logger,
        resource_id: Optional[Union[List[str], str]],
        server_url: str,
        extra_context_to_return: Optional[Dict[str, Any]] = None,
        parameters: FhirReceiverParameters,
        page_number: Optional[int] = None,
        page_size: Optional[int] = None,
        last_updated_after: Optional[datetime] = None,
        last_updated_before: Optional[datetime] = None,
        data_chunk_handler: Optional[HandleStreamingChunkFunction] = None,
    ) -> AsyncGenerator[FhirGetResponse, None]:
        """
        Sends a fhir request to the fhir client sdk


        :param logger:
        :param resource_id:
        :param server_url:
        :param parameters:
        :param page_number: page number to query
        :param page_size: page size to query
        :param last_updated_after: last updated after date
        :param last_updated_before: last updated before date
        :param extra_context_to_return: a dict to return with every row (separate_bundle_resources is set)
                                        or with FhirGetResponse
        :param data_chunk_handler: function to handle the data chunk
        :return:
        :rtype:
        """

        fhir_client: FhirClient = get_fhir_client(
            logger=logger,
            server_url=server_url,
            auth_server_url=parameters.auth_server_url,
            auth_client_id=parameters.auth_client_id,
            auth_client_secret=parameters.auth_client_secret,
            auth_login_token=parameters.auth_login_token,
            auth_access_token=parameters.auth_access_token,
            auth_scopes=parameters.auth_scopes,
            log_level=parameters.log_level,
            auth_well_known_url=parameters.auth_well_known_url,
        )
        if (
            not parameters.graph_json
        ):  # graph_json passes this in the simulate_graph_async()
            fhir_client = fhir_client.separate_bundle_resources(
                parameters.separate_bundle_resources
            )
        fhir_client = fhir_client.expand_fhir_bundle(parameters.expand_fhir_bundle)
        if parameters.accept_type is not None:
            fhir_client = fhir_client.accept(parameters.accept_type)
        if parameters.content_type is not None:
            fhir_client = fhir_client.content_type(parameters.content_type)
        if parameters.accept_encoding is not None:
            fhir_client = fhir_client.accept_encoding(parameters.accept_encoding)
        if parameters.additional_request_headers is not None:
            logger.info(
                f"Additional Request Headers to be sent - {parameters.additional_request_headers}"
            )
            fhir_client = fhir_client.additional_request_headers(
                parameters.additional_request_headers
            )

        fhir_client = fhir_client.resource(parameters.resource_type)
        if resource_id:
            fhir_client = fhir_client.id_(resource_id)
            if parameters.filter_by_resource:
                fhir_client = fhir_client.filter_by_resource(
                    parameters.filter_by_resource
                )
                if parameters.filter_parameter:
                    fhir_client = fhir_client.filter_parameter(
                        parameters.filter_parameter
                    )
        # add action to url
        if parameters.action:
            fhir_client = fhir_client.action(parameters.action)
            if parameters.action_payload:
                fhir_client = fhir_client.action_payload(parameters.action_payload)
        # add a query for just desired properties
        if parameters.include_only_properties:
            fhir_client = fhir_client.include_only_properties(
                include_only_properties=parameters.include_only_properties
            )
        if page_size and page_number is not None:
            fhir_client = fhir_client.page_size(page_size).page_number(page_number)
        if parameters.sort_fields is not None:
            fhir_client = fhir_client.sort_fields(parameters.sort_fields)

        if parameters.additional_parameters:
            fhir_client = fhir_client.additional_parameters(
                parameters.additional_parameters
            )

        # have to done here since this arg can be used twice
        if last_updated_before:
            fhir_client = fhir_client.last_updated_before(last_updated_before)
        if last_updated_after:
            fhir_client = fhir_client.last_updated_after(last_updated_after)

        if extra_context_to_return:
            fhir_client = fhir_client.extra_context_to_return(extra_context_to_return)

        if parameters.retry_count is not None:
            fhir_client = fhir_client.retry_count(parameters.retry_count)

        if parameters.exclude_status_codes_from_retry:
            fhir_client = fhir_client.exclude_status_codes_from_retry(
                parameters.exclude_status_codes_from_retry
            )

        if parameters.limit is not None:
            fhir_client = fhir_client.limit(limit=parameters.limit)

        if parameters.use_data_streaming:
            fhir_client = fhir_client.use_data_streaming(parameters.use_data_streaming)

        if parameters.refresh_token_function:
            fhir_client = fhir_client.refresh_token_function(
                parameters.refresh_token_function
            )

        if parameters.graph_json:
            async for r in fhir_client.simulate_graph_streaming_async(
                id_=(
                    cast(str, resource_id)
                    if not isinstance(resource_id, list)
                    else resource_id[0]
                ),
                graph_json=parameters.graph_json,
                contained=False,
                separate_bundle_resources=parameters.separate_bundle_resources,
            ):
                yield r
        else:
            async for r in fhir_client.get_streaming_async(
                data_chunk_handler=data_chunk_handler
            ):
                yield r

    @staticmethod
    async def get_batch_results_paging_async(
        *,
        last_updated_after: Optional[datetime],
        last_updated_before: Optional[datetime],
        limit: Optional[int],
        page_size: Optional[int],
        parameters: FhirReceiverParameters,
        server_url: Optional[str],
    ) -> AsyncGenerator[GetBatchResult, None]:
        assert server_url
        assert (
            not parameters.use_data_streaming
        ), f"Data streaming is not supported with paging.  Use get_all_resources_async() instead."
        additional_parameters: Optional[List[str]] = parameters.additional_parameters
        if not page_size:
            page_size = limit
        # if paging is requested then iterate through the pages until the response is empty
        page_number: int = 0
        server_page_number: int = 0
        has_next_page: bool = True
        loop_number: int = 0
        logger: Logger = get_logger(
            __name__,
            level=(
                parameters.log_level if parameters and parameters.log_level else "INFO"
            ),
        )
        while has_next_page:
            loop_number += 1
            async for result in FhirReceiverProcessor.send_fhir_request_async(
                logger=logger,
                resource_id=None,
                server_url=server_url,
                page_number=server_page_number,  # since we're setting id:above we can leave this as 0
                page_size=page_size,
                last_updated_after=last_updated_after,
                last_updated_before=last_updated_before,
                parameters=parameters.clone()
                .set_additional_parameters(additional_parameters)
                .set_expand_fhir_bundle(
                    True
                ),  # always expand bundles so we can aggregate resources properly.
                # we'll convert back into a bundle at the end if necessary
            ):
                resources: List[str] = []
                errors: List[GetBatchError] = []

                result_response: List[str] = []
                try:
                    batch_result: GetBatchResult = (
                        FhirReceiverProcessor.read_resources_and_errors_from_response(
                            response=result
                        )
                    )
                    result_response = batch_result.resources
                    errors = batch_result.errors
                except JSONDecodeError as e:
                    if parameters.error_view:
                        errors.append(
                            GetBatchError(
                                url=result.url,
                                status_code=result.status,
                                error_text=str(e) + " : " + result.responses,
                                request_id=result.request_id,
                            )
                        )
                    else:
                        raise FhirParserException(
                            url=result.url,
                            message="Parsing result as json failed",
                            json_data=result.responses,
                            response_status_code=result.status,
                            request_id=result.request_id,
                        ) from e

                if len(result_response) > 0:
                    # get id of last resource
                    json_resources: List[Dict[str, Any]] = [
                        json.loads(r) for r in result_response
                    ]
                    if isinstance(json_resources, list):  # normal response
                        if len(json_resources) > 0:  # received any resources back
                            last_json_resource = json_resources[-1]
                            if result.next_url:
                                # if server has sent back a next url then use that
                                next_url: Optional[str] = result.next_url
                                if next_url:
                                    next_uri: furl = furl(next_url)
                                    additional_parameters = [
                                        f"{k}={v}"
                                        for k, v in next_uri.query.params.allitems()
                                    ]
                                    # remove any entry for id:above
                                    additional_parameters = list(
                                        filter(
                                            lambda x: not x.startswith("_count")
                                            and not x.startswith("_element"),
                                            additional_parameters,
                                        )
                                    )
                            elif (
                                "id" in last_json_resource
                                and parameters.use_id_above_for_paging
                            ):
                                # use id:above to optimize the next query
                                id_of_last_resource = last_json_resource["id"]
                                if not additional_parameters:
                                    additional_parameters = []
                                # remove any entry for id:above
                                additional_parameters = list(
                                    filter(
                                        lambda x: not x.startswith("id:above"),
                                        additional_parameters,
                                    )
                                )
                                additional_parameters.append(
                                    f"id:above={id_of_last_resource}"
                                )
                            else:
                                server_page_number += 1
                            resources = resources + result_response
                        page_number += 1
                        if limit and 0 < limit <= len(resources):
                            has_next_page = False
                elif result.status == 200:
                    # no resources returned but status is 200 so we're done
                    has_next_page = False
                else:
                    if result.status == 404 and loop_number > 1:
                        # 404 (not found) is fine since it just means we ran out of data while paging
                        pass
                    elif result.status not in parameters.ignore_status_codes:
                        raise FhirReceiverException(
                            url=result.url,
                            json_data=result.responses,
                            response_text=result.responses,
                            response_status_code=result.status,
                            message=(
                                "Error received from server"
                                if len(errors) == 0
                                else "\n".join([e.error_text for e in errors])
                            ),
                            request_id=result.request_id,
                        )
                    has_next_page = False

                # This is the right thing to do but because of backward-compatibility we can't do it
                # we fix all the pipelines that rely on this returning resources instead of bundle
                # if parameters.expand_fhir_bundle is False:
                #     # convert the resources back into a bundle
                #     bundle: Dict[str, Any] = {
                #         "resourceType": "Bundle",
                #         "type": "searchset",
                #         "entry": [{"resource": json.loads(r)} for r in resources],
                #     }
                #     resources = [json.dumps(bundle)]

                # Now return the data back to the caller
                yield GetBatchResult(resources=resources, errors=errors)

    @staticmethod
    async def get_batch_result_streaming_async(
        *,
        last_updated_after: Optional[datetime],
        last_updated_before: Optional[datetime],
        parameters: FhirReceiverParameters,
        server_url: Optional[str],
        limit: Optional[int] = None,
    ) -> AsyncGenerator[Dict[str, Any], None]:
        assert server_url
        result: FhirGetResponse
        logger: Logger = get_logger(
            __name__,
            level=(
                parameters.log_level if parameters and parameters.log_level else "INFO"
            ),
        )
        # Iterate till we have next page. We are using cursor based pagination here so next page will have records
        # with id greater than previous page last record id.
        has_next_page: bool = True
        additional_parameters = parameters.additional_parameters or []
        count: int = 0
        while has_next_page:
            async for result in FhirReceiverProcessor.send_fhir_request_async(
                logger=logger,
                resource_id=None,
                server_url=server_url,
                page_size=limit,
                last_updated_after=last_updated_after,
                last_updated_before=last_updated_before,
                parameters=parameters.clone().set_additional_parameters(
                    additional_parameters
                ),
            ):
                resources: List[str] = []
                errors: List[GetBatchError] = []
                result_response: List[str] = []
                try:
                    batch_result1: GetBatchResult = (
                        FhirReceiverProcessor.read_resources_and_errors_from_response(
                            response=result
                        )
                    )
                    result_response = batch_result1.resources
                    errors = batch_result1.errors
                except JSONDecodeError as e:
                    if parameters.error_view:
                        errors.append(
                            GetBatchError(
                                url=result.url,
                                status_code=result.status,
                                error_text=str(e) + " : " + result.responses,
                                request_id=result.request_id,
                            )
                        )
                    else:
                        raise FhirParserException(
                            url=result.url,
                            message="Parsing result as json failed",
                            json_data=result.responses,
                            response_status_code=result.status,
                            request_id=result.request_id,
                        ) from e
                if len(result_response) > 0:
                    resources = resources + result_response
                    count += len(resources)
                    # get id of last resource
                    json_resources: List[Dict[str, Any]] = [
                        json.loads(r) for r in result_response
                    ]
                    if isinstance(json_resources, list):
                        if len(json_resources) > 0:  # received any resources back
                            last_json_resource = json_resources[-1]
                            id_of_last_resource = None
                            if parameters.use_uuid_for_id_above:
                                for identifier in last_json_resource.get(
                                    "identifier", []
                                ):
                                    if identifier.get("id") == "uuid":
                                        id_of_last_resource = identifier.get("value")
                            elif "id" in last_json_resource:
                                id_of_last_resource = last_json_resource["id"]
                            if id_of_last_resource:
                                # use id:above to optimize the next query and remove any entry for id:above
                                additional_parameters = list(
                                    filter(
                                        lambda x: not x.startswith("id:above"),
                                        additional_parameters,
                                    )
                                )
                                additional_parameters.append(
                                    f"id:above={id_of_last_resource}"
                                )
                            else:
                                has_next_page = False
                            if limit and 0 < limit <= count:
                                has_next_page = False
                elif result.status == 200:
                    # no resources returned but status is 200, so we're done
                    has_next_page = False
                else:
                    if result.status == 404:
                        yield GetBatchResult(resources=[], errors=[]).to_dict()
                    elif result.status not in parameters.ignore_status_codes:
                        raise FhirReceiverException(
                            url=result.url,
                            json_data=result.responses,
                            response_text=result.responses,
                            response_status_code=result.status,
                            message=(
                                "Error received from server"
                                if len(errors) == 0
                                else "\n".join([e.error_text for e in errors])
                            ),
                            request_id=result.request_id,
                        )
                    has_next_page = False

                yield GetBatchResult(resources=resources, errors=errors).to_dict()

    @staticmethod
    def read_resources_and_errors_from_response(
        response: FhirGetResponse,
    ) -> GetBatchResult:
        all_resources: List[Dict[str, Any]] = response.get_resources()
        resources_except_operation_outcomes: List[Dict[str, Any]] = [
            r for r in all_resources if r.get("resourceType") != "OperationOutcome"
        ]
        operation_outcomes: List[Dict[str, Any]] = [
            r for r in all_resources if r.get("resourceType") == "OperationOutcome"
        ]

        errors: List[GetBatchError] = [
            GetBatchError(
                request_id=response.request_id,
                url=response.url,
                status_code=response.status,
                error_text=json.dumps(o, indent=2),
            )
            for o in operation_outcomes
        ]
        return GetBatchResult(
            resources=[json.dumps(r) for r in resources_except_operation_outcomes],
            errors=errors,
        )



class FhirGetResponseSchema:
    """
    This class provides names for columns in FhirGetResponse

    Should match to
    https://github.com/icanbwell/helix.fhir.client.sdk/blob/main/helix_fhir_client_sdk/responses/fhir_get_response.py
    """

    partition_index = "partition_index"
    sent = "sent"
    received = "received"
    responses = "responses"
    first = "first"
    last = "last"
    error_text = "error_text"
    url = "url"
    status_code = "status_code"
    request_id = "request_id"
    access_token = "access_token"
    extra_context_to_return = "extra_context_to_return"



import asyncio

import pytest

from helixcore.utilities.async_safe_buffer.v1.async_safe_buffer import (
    AsyncSafeBuffer,
)


@pytest.fixture
async def buffer() -> AsyncSafeBuffer[str, int]:
    return AsyncSafeBuffer[str, int]()


@pytest.mark.asyncio
async def test_empty_buffer_initialization(buffer: AsyncSafeBuffer[str, int]) -> None:
    assert await buffer.total_items() == 0
    assert await buffer.items_by_type() == {}
    assert await buffer.get_up_to(10) == {}


@pytest.mark.asyncio
async def test_add_single_item(buffer: AsyncSafeBuffer[str, int]) -> None:
    await buffer.add("metric1", 42)

    assert await buffer.total_items() == 1
    assert await buffer.items_by_type() == {"metric1": 1}

    result = await buffer.get_up_to(1)
    assert result == {"metric1": [42]}


@pytest.mark.asyncio
async def test_add_list_items(buffer: AsyncSafeBuffer[str, int]) -> None:
    items = [1, 2, 3]
    await buffer.add_list([("metric1", item) for item in items])

    assert await buffer.total_items() == 3
    assert await buffer.items_by_type() == {"metric1": 3}

    result = await buffer.get_up_to(3)
    assert result == {"metric1": [1, 2, 3]}


@pytest.mark.asyncio
async def test_get_up_to_partial(buffer: AsyncSafeBuffer[str, int]) -> None:
    items = [1, 2, 3]
    await buffer.add_list([("metric1", item) for item in items])
    result = await buffer.get_up_to(2)

    assert len(result["metric1"]) == 2
    assert await buffer.total_items() == 1


@pytest.mark.asyncio
async def test_multiple_metric_types(buffer: AsyncSafeBuffer[str, int]) -> None:
    await buffer.add("metric1", 1)
    await buffer.add("metric2", 2)

    assert await buffer.total_items() == 2
    assert await buffer.items_by_type() == {"metric1": 1, "metric2": 1}

    result = await buffer.get_up_to(2)
    assert "metric1" in result
    assert "metric2" in result
    assert len(result["metric1"]) == 1
    assert len(result["metric2"]) == 1


@pytest.mark.asyncio
async def test_clear_buffer(buffer: AsyncSafeBuffer[str, int]) -> None:
    await buffer.add("metric1", 1)
    await buffer.add("metric2", 2)
    await buffer.clear()

    assert await buffer.total_items() == 0
    assert await buffer.items_by_type() == {}


@pytest.mark.asyncio
async def test_empty_metric_type_removal(buffer: AsyncSafeBuffer[str, int]) -> None:
    await buffer.add("metric1", 1)
    result = await buffer.get_up_to(1)

    assert {"metric1": 0} == await buffer.items_by_type()


@pytest.mark.asyncio
async def test_concurrent_access() -> None:
    buffer: AsyncSafeBuffer[str, int] = AsyncSafeBuffer[str, int]()

    async def add_items() -> None:
        for i in range(100):
            await buffer.add("metric1", i)

    async def get_items() -> None:
        for _ in range(10):
            await buffer.get_up_to(10)

    # Create multiple tasks
    tasks = [add_items(), add_items(), get_items(), get_items()]

    # Run tasks concurrently
    await asyncio.gather(*tasks)

    # Verify final state
    total = await buffer.total_items()
    assert total >= 0  # The exact number will depend on timing


@pytest.mark.asyncio
async def test_different_types() -> None:
    buffer: AsyncSafeBuffer[str, str] = AsyncSafeBuffer[str, str]()
    await buffer.add("metric1", "value1")
    await buffer.add_list([("metric2", "value2"), ("metric2", "value3")])

    result = await buffer.get_up_to(3)
    assert isinstance(result["metric1"][0], str)
    assert isinstance(result["metric2"][0], str)


@pytest.mark.asyncio
async def test_get_up_to_with_empty_buffer(buffer: AsyncSafeBuffer[str, int]) -> None:
    result = await buffer.get_up_to(10)
    assert result == {}



from abc import ABC, abstractmethod
from logging import Logger
from types import TracebackType
from typing import Any, Dict, List, Type, Sequence, Optional

from helixcore.utilities.metrics.base_metrics import BaseMetric


class BaseMetricsWriter(ABC):
    def __init__(
        self,
        *,
        schema_name: str,
        logger: Optional[Logger],
        metric_table_map: Dict[str, Optional[str]],
    ) -> None:
        self.schema_name: str = schema_name
        self.logger: Optional[Logger] = logger
        self.metric_table_map: Dict[str, Optional[str]] = metric_table_map
        self.tables_created_for_metric: Dict[str, bool] = {}
        self.has_database_been_created: bool = False

    @abstractmethod
    def __enter__(self) -> "BaseMetricsWriter":
        pass

    @abstractmethod
    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        pass

    @abstractmethod
    def create_table_if_not_exists(self, *, metric: BaseMetric) -> None:
        pass

    @abstractmethod
    def write_single_metric_to_table(self, *, metric: BaseMetric) -> Optional[int]:
        pass

    @abstractmethod
    def write_metrics_to_table(self, *, metrics: Sequence[BaseMetric]) -> Optional[int]:
        pass

    @abstractmethod
    def read_metrics_from_table(self, metric: BaseMetric) -> List[Dict[str, Any]]:
        pass



import asyncio
from asyncio import Lock
from collections import deque
from typing import Dict, List, Tuple


class AsyncSafeBuffer[TKey, TValue]:
    def __init__(self) -> None:
        self._lock: Lock = asyncio.Lock()
        self._buffer: Dict[TKey, deque[TValue]] = {}
        self._total_items: int = 0

    async def add(self, metric_type: TKey, item: TValue) -> None:
        """Add an item to the buffer."""
        async with self._lock:
            if metric_type not in self._buffer:
                self._buffer[metric_type] = deque()
            self._buffer[metric_type].append(item)
            self._total_items += 1

    async def add_list(self, items: List[Tuple[TKey, TValue]]) -> None:
        """Add a list of items to the buffer."""
        async with self._lock:
            metric_type: TKey
            item: TValue
            for metric_type, item in items:
                if metric_type not in self._buffer:
                    self._buffer[metric_type] = deque()
                self._buffer[metric_type].append(item)
                self._total_items += 1

    async def get_all(self) -> Dict[TKey, List[TValue]]:
        """Get all items from the buffer."""
        async with self._lock:
            result = {k: list(v) for k, v in self._buffer.items()}
            self._buffer.clear()
            self._total_items = 0
            return result

    async def get_up_to(self, count: int) -> Dict[TKey, List[TValue]]:
        """
        Get up to 'count' items from the buffer.
        Returns a tuple of (items_dict, items_retrieved)
        """
        async with self._lock:
            result: Dict[TKey, List[TValue]] = {}
            items_retrieved = 0

            if self._total_items == 0:
                return result

            # Copy items up to count, maintaining proportion across types
            remaining_count = count
            while remaining_count > 0 and self._total_items > 0:
                # Get one item from each non-empty buffer type
                for metric_type in list(self._buffer.keys()):
                    if not self._buffer[metric_type]:
                        del self._buffer[metric_type]
                        continue

                    if metric_type not in result:
                        result[metric_type] = []

                    item = self._buffer[metric_type].popleft()
                    result[metric_type].append(item)
                    items_retrieved += 1
                    self._total_items -= 1
                    remaining_count -= 1

                    if remaining_count <= 0:
                        break

            return result

    async def clear(self) -> None:
        """Clear the buffer."""
        async with self._lock:
            self._buffer.clear()
            self._total_items = 0

    async def total_items(self) -> int:
        """Get the total number of items in the buffer."""
        async with self._lock:
            return self._total_items

    async def total_items_by_type(self, metric_type: TKey) -> int:
        """Get the total number of items in the buffer for specified type."""
        async with self._lock:
            return len(self._buffer.get(metric_type, []))

    async def items_by_type(self) -> Dict[TKey, int]:
        """Get the count of items for each type."""
        async with self._lock:
            return {k: len(v) for k, v in self._buffer.items()}



class FhirMergeResponseItemSchema:
    updated = "updated"
    created = "created"
    deleted = "deleted"
    issue = "issue"
    id_ = "id"
    uuid = "uuid"
    resourceType = "resourceType"
    sourceAssigningAuthority = "sourceAssigningAuthority"
    resource_version = "resource_version"
    message = "message"
    error = "error"
    token = "token"
    resource_json = "resource_json"



from typing import Any, Dict, Optional, Callable, Type

from helixcore.utilities.telemetry.telemetry import (
    Telemetry,
)
from helixcore.utilities.telemetry.telemetry_context import (
    TelemetryContext,
)
from helixcore.utilities.telemetry.telemetry_span_creator import (
    TelemetrySpanCreator,
)


class TelemetryFactory:

    _registry: Dict[str, type[Telemetry]] = {}

    def __init__(self, *, telemetry_context: TelemetryContext) -> None:
        """
        Telemetry factory used to create telemetry instances based on the telemetry context


        :param telemetry_context: telemetry context
        """
        assert telemetry_context is not None
        assert telemetry_context.telemetry_provider is not None
        self.telemetry_context: TelemetryContext = telemetry_context

    @classmethod
    def register_telemetry_class(
        cls, *, name: str, telemetry_class: Type[Telemetry]
    ) -> None:
        registration_name = name

        # Validate that the class is a Telemetry subclass
        if not issubclass(telemetry_class, Telemetry):
            raise TypeError(
                f"{telemetry_class.__name__} must be a subclass of Telemetry"
            )

        # Register the class in the factory's registry
        cls._registry[registration_name] = telemetry_class

    @classmethod
    def register_telemetry(
        cls, name: str
    ) -> Callable[[Type[Telemetry]], Type[Telemetry]]:
        """
        Decorator to register Telemetry subclasses in the factory registry.

        Args:
            name (Optional[str], optional): Custom registration name.
                Defaults to the class name if not provided.

        Returns:
            Callable: A decorator function for registering Telemetry classes
        """

        def decorator(telemetry_class: Type[Telemetry]) -> Type[Telemetry]:
            # Use provided name or fallback to class name
            registration_name = name

            # Validate that the class is a Telemetry subclass
            if not issubclass(telemetry_class, Telemetry):
                raise TypeError(
                    f"{telemetry_class.__name__} must be a subclass of Telemetry"
                )

            # Register the class in the factory's registry
            cls._registry[registration_name] = telemetry_class

            return telemetry_class

        return decorator

    def create(self, *, log_level: Optional[str | int]) -> Telemetry:
        """
        Create a telemetry instance

        :return: telemetry instance
        """
        assert (
            self.telemetry_context.telemetry_provider in self._registry
        ), f"Telemetry {self.telemetry_context.telemetry_provider} not found in registry.  Did you register a class for it using register_telemetry_class()?"
        return self._registry[self.telemetry_context.telemetry_provider](
            telemetry_context=self.telemetry_context, log_level=log_level
        )

    def create_telemetry_span_creator(
        self, *, log_level: Optional[str | int]
    ) -> TelemetrySpanCreator:
        """
        Create a telemetry span creator

        :return: telemetry span creator
        """
        return TelemetrySpanCreator(
            telemetry=self.create(log_level=log_level),
            telemetry_context=self.telemetry_context,
        )

    def __getstate__(self) -> Dict[str, Any]:
        assert "TelemetryFactory should not be pickled (serialized) as it is not serializable"
        # Exclude certain properties from being pickled otherwise they cause errors in pickling
        return {
            k: v
            for k, v in self.__dict__.items()
            if k not in ["_telemetry_factory", "_telemetry"]
        }



from datetime import datetime
from types import TracebackType
from typing import Optional

from helixcore.progress_logger.progress_logger import ProgressLogger


class ProgressLogMetric:
    def __init__(
        self,
        name: str,
        progress_logger: Optional[ProgressLogger],
        measure_in_seconds: Optional[bool] = None,
    ):
        self.progress_logger: Optional[ProgressLogger] = progress_logger
        self.name: str = name
        self.start_time: datetime = datetime.now()
        self.measure_in_seconds: Optional[bool] = measure_in_seconds

    def __enter__(self) -> "ProgressLogMetric":
        return self.start()

    def __exit__(
        self,
        exc_type: Optional[BaseException],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.stop()

    def start(self) -> "ProgressLogMetric":
        """
        start
        :return:
        """
        self.start_time = datetime.now()
        return self

    def stop(self) -> None:
        """
        stop
        """
        if self.name and self.progress_logger:
            end_time: datetime = datetime.now()
            seconds = (end_time - self.start_time).total_seconds()
            if self.progress_logger:
                if self.measure_in_seconds:
                    self.progress_logger.log_metric(self.name, seconds)
                else:
                    time_diff_in_minutes: float = seconds / 60
                    self.progress_logger.log_metric(self.name, time_diff_in_minutes)



from logging import Logger
from typing import Optional

from helixcore.utilities.telemetry.telemetry_span_creator import (
    TelemetrySpanCreator,
)
from helixcore.utilities.metrics.writer.base_metrics_writer_async import (
    BaseMetricsWriterAsync,
)
from helixcore.utilities.metrics.writer.base_metrics_writer_parameters import (
    BaseMetricsWriterParameters,
)
from helixcore.utilities.metrics.writer.v2.metrics_writer import (
    MetricsWriter,
)
from helixcore.utilities.metrics.writer.v2.metrics_writer_parallel import (
    MetricsWriterParallel,
)


class MetricsWriterFactory:
    def __init__(
        self,
        *,
        parameters: BaseMetricsWriterParameters,
        logger: Optional[Logger],
    ) -> None:
        """
        This class creates a metrics writer


        :param parameters: parameters for the metrics writer
        :param logger: logger to use
        """

        assert parameters is not None, "parameters should not be None"
        assert isinstance(
            parameters, BaseMetricsWriterParameters
        ), "parameters should be an instance of BaseMetricsWriterParameters"

        self.logger: Optional[Logger] = logger

        self.parameters: BaseMetricsWriterParameters = parameters

    def create_metrics_writer(
        self, *, telemetry_span_creator: TelemetrySpanCreator
    ) -> BaseMetricsWriterAsync:
        """
        Creates a metrics writer

        :return: metrics writer
        """
        return (
            MetricsWriter(
                logger=self.logger,
                parameters=self.parameters,
                telemetry_span_creator=telemetry_span_creator,
            )
            if not self.parameters.buffer_length
            else MetricsWriterParallel(
                logger=self.logger,
                parameters=self.parameters,
                telemetry_span_creator=telemetry_span_creator,
            )
        )



import re
from os import environ
from pathlib import Path
from tempfile import TemporaryDirectory
from types import TracebackType
from typing import Optional, List, Dict, Any, cast

# noinspection PyPackageRequirements
import mlflow

# noinspection PyPackageRequirements
from mlflow.entities import Experiment, RunStatus, Run
from mlflow.store.tracking.file_store import FileStore

from helixcore.event_loggers.event_logger import EventLogger
from helixcore.logger.log_level import LogLevel
from helixcore.logger.yarn_logger import get_logger


class MlFlowConfig:
    def __init__(
        self,
        mlflow_tracking_url: str,
        artifact_url: str,
        experiment_name: str,
        flow_run_name: str,
        parameters: Dict[str, Any],
    ):
        self.mlflow_tracking_url = mlflow_tracking_url
        self.artifact_url = artifact_url
        self.experiment_name = experiment_name
        self.flow_run_name = flow_run_name
        self.parameters = parameters

    def clone(self) -> "MlFlowConfig":
        return MlFlowConfig(
            mlflow_tracking_url=self.mlflow_tracking_url,
            artifact_url=self.artifact_url,
            experiment_name=self.experiment_name,
            flow_run_name=self.flow_run_name,
            parameters=self.parameters.copy(),
        )


class ProgressLogger:
    def __init__(
        self,
        event_loggers: Optional[List[EventLogger]] = None,
        mlflow_config: Optional[MlFlowConfig] = None,
    ) -> None:
        self.logger = get_logger(__name__)
        self.event_loggers: Optional[List[EventLogger]] = event_loggers
        self.mlflow_config: Optional[MlFlowConfig] = mlflow_config
        system_log_level_text: Optional[str] = environ.get("LOGLEVEL")
        self.system_log_level: Optional[LogLevel] = (
            LogLevel.from_str(system_log_level_text)
            if system_log_level_text is not None
            else None
        )

    def __enter__(self) -> "ProgressLogger":
        if self.mlflow_config is None:
            self.logger.debug("MLFLOW IS NOT ENABLED")
            return self
        self.logger.debug("MLFLOW IS ENABLED")

        try:
            mlflow.set_tracking_uri(self.mlflow_config.mlflow_tracking_url)
            self.logger.debug(f"MLFLOW TRACKING URL: {mlflow.get_tracking_uri()}")
            self.start_mlflow_run(
                run_name=self.mlflow_config.flow_run_name, is_nested=False
            )
            # set the parameters used in the pipeline run
            self.log_params(params=self.mlflow_config.parameters)
        except Exception as e:
            self.log_event(
                "mlflow initialization error. suppressing and continuing.",
                str({e}),
                log_level=LogLevel.ERROR,
            )
            self.mlflow_config = None  # since there are checks for this everywhere, quick way to bypass MLFlow

        return self

    def __exit__(
        self,
        exc_type: Optional[BaseException],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.logger.debug("ENDING PARENT RUN")
        if exc_value:
            # there was an exception so mark the parent run as failed
            self.end_mlflow_run(status=RunStatus.FAILED)
        else:
            # safe to call without checking if we have a tracking url set for mlflow
            self.end_mlflow_run()

    def start_mlflow_run(self, run_name: str, is_nested: bool = True) -> None:
        if self.mlflow_config is None:
            return
        # get or create experiment
        experiment: Optional[Experiment] = mlflow.get_experiment_by_name(
            name=self.mlflow_config.experiment_name
        )

        if experiment is None:
            experiment_id: str = mlflow.create_experiment(
                name=self.mlflow_config.experiment_name,
                artifact_location=self.mlflow_config.artifact_url,
            )
        else:
            experiment_id = experiment.experiment_id
        mlflow.set_experiment(experiment_id=experiment_id)

        mlflow.start_run(run_name=run_name, nested=is_nested)

    # noinspection PyMethodMayBeStatic
    def end_mlflow_run(self, status: RunStatus = RunStatus.FINISHED) -> None:
        if self.mlflow_config is None:
            return
        mlflow.end_run(
            status=RunStatus.to_string(status)  # type:ignore[no-untyped-call]
        )

    def log_metric(
        self,
        name: str,
        time_diff_in_minutes: float,
        log_level: LogLevel = LogLevel.TRACE,
    ) -> None:
        self.logger.debug(f"{name}: {time_diff_in_minutes} min")
        if self.mlflow_config is not None:
            try:
                mlflow.log_metric(
                    key=self.__mlflow_clean_string(name), value=time_diff_in_minutes
                )
            except Exception as e:
                self.log_event("mlflow log metric error", str({e}), log_level=log_level)

    def log_param(
        self, key: str, value: str, log_level: LogLevel = LogLevel.TRACE
    ) -> None:
        self.write_to_log(name=key, message=value)
        if self.mlflow_config is not None:
            try:
                mlflow.log_param(
                    key=self.__mlflow_clean_string(key),
                    value=self.__mlflow_clean_param_value(value),
                )
            except Exception as e:
                self.log_event("mlflow log param error", str({e}), log_level=log_level)

    def log_params(
        self, params: Dict[str, Any], log_level: LogLevel = LogLevel.TRACE
    ) -> None:
        if self.mlflow_config is not None:
            # intentionally not using mlflow.log_params due to issues with its SqlAlchemy implementation
            for key, value in params.items():
                self.log_param(key=key, value=value, log_level=log_level)

    @staticmethod
    def __mlflow_clean_string(value: str) -> str:
        """

        MLFlow keys may only contain alphanumerics, underscores (_),
        dashes (-), periods (.), spaces ( ), and slashes (/)


        mlflow run metric names through https://docs.python.org/3/library/os.path.html#os.path.normpath when validating
        metric names
        (https://github.com/mlflow/mlflow/blob/217799b10780b22f787137f80f5cf5c2b5cf85b1/mlflow/utils/validation.py#L95).
        one side effect of this is if the value contains `//` it will be changed to `/`
        and fail the _validate_metric_name check.
        """
        value = str(value).replace("//", "/")
        # noinspection RegExpRedundantEscape
        return re.sub(r"[^\w\-\.\s\/]", "-", value)

    @staticmethod
    def __mlflow_clean_param_value(param_value: str) -> str:
        """
        replace sensitive values in the string with asterisks
        """
        sensitive_value_replacement: str = "*******"
        db_password_regex = r"(?<=:)\w+(?=@)"
        cleaned_value = re.sub(
            db_password_regex, sensitive_value_replacement, str(param_value)
        )

        return cleaned_value

    # noinspection PyUnusedLocal
    def log_artifact(
        self, key: str, contents: str, folder_path: Optional[str] = None
    ) -> None:
        if self.mlflow_config is not None:
            try:
                with TemporaryDirectory() as temp_dir_name:
                    data_dir: Path = Path(temp_dir_name)
                    file_path: Path = data_dir.joinpath(key)
                    file_path.parent.mkdir(exist_ok=True, parents=True)
                    with open(file_path, "w") as file:
                        file.write(contents)
                        self.logger.debug(f"Wrote sql to {file_path}")

                    if self.mlflow_config is not None:
                        mlflow.log_artifact(local_path=str(file_path))

            except Exception as e:
                self.log_event("Error in log_artifact writing to mlflow", str(e))

    def write_to_log(
        self, name: str, message: str = "", log_level: LogLevel = LogLevel.INFO
    ) -> bool:
        if log_level == LogLevel.ERROR:
            self.logger.error(name + ": " + str(message))
        elif log_level == LogLevel.INFO:
            self.logger.info(name + ": " + str(message))
        else:
            self.logger.debug(name + ": " + str(message))
        return True

    def log_exception(self, event_name: str, event_text: str, ex: Exception) -> None:
        self.log_artifact("_exception.txt", str(ex))
        if self.event_loggers:
            for event_logger in self.event_loggers:
                event_logger.log_exception(
                    event_name=event_name, event_text=event_text, ex=ex
                )

    def log_progress_event(
        self,
        event_name: str,
        current: int,
        total: int,
        event_format_string: str,
        backoff: bool = True,
        log_level: LogLevel = LogLevel.TRACE,
    ) -> None:
        self.logger.debug(event_format_string.format(event_name, current, total))
        if not self.system_log_level or self.system_log_level == LogLevel.INFO:
            if (
                log_level == LogLevel.INFO or log_level == LogLevel.ERROR
            ):  # log only INFO messages
                if self.event_loggers:
                    for event_logger in self.event_loggers:
                        event_logger.log_progress_event(
                            event_name=event_name,
                            current=current,
                            total=total,
                            event_format_string=event_format_string,
                            backoff=backoff,
                        )
        else:  # LOGLEVEL is lower than INFO
            if self.event_loggers:
                for event_logger in self.event_loggers:
                    event_logger.log_progress_event(
                        event_name=event_name,
                        current=current,
                        total=total,
                        event_format_string=event_format_string,
                        backoff=backoff,
                    )

    def log_event(
        self, event_name: str, event_text: str, log_level: LogLevel = LogLevel.TRACE
    ) -> None:
        self.write_to_log(name=event_name, message=event_text)
        if not self.system_log_level or self.system_log_level == LogLevel.INFO:
            if (
                log_level == LogLevel.INFO or log_level == LogLevel.ERROR
            ):  # log only INFO messages
                if self.event_loggers:
                    for event_logger in self.event_loggers:
                        event_logger.log_event(
                            event_name=event_name, event_text=event_text
                        )
        else:
            if self.event_loggers:
                for event_logger in self.event_loggers:
                    event_logger.log_event(event_name=event_name, event_text=event_text)

    @staticmethod
    def clean_experiments(*, tracking_uri: Optional[str]) -> None:

        logger = get_logger(__name__)
        if tracking_uri is None:
            return
        mlflow.set_tracking_uri(tracking_uri)
        # List all experiments
        experiments = ProgressLogger.get_experiments()
        # Delete each experiment
        for exp in experiments:
            if exp.experiment_id == FileStore.DEFAULT_EXPERIMENT_ID:
                continue

            assert isinstance(
                exp, Experiment
            ), f"Expected Experiment, got {type(exp)}: {exp}"
            assert isinstance(
                exp.experiment_id, str
            ), f"Expected str, got {type(exp.experiment_id)}: {exp.experiment_id}"
            # Get all runs in the experiment
            runs: List[Run] = cast(
                List[Run],
                mlflow.search_runs(
                    experiment_ids=[exp.experiment_id], output_format="list"
                ),
            )

            # Delete each run
            run: Run
            for run in runs:
                assert isinstance(run, Run), f"Expected Run, got {type(run)}: {run}"
                mlflow.delete_run(run.info.run_id)
                logger.info(f"Run with ID {run.info.run_id} has been deleted")

            mlflow.delete_experiment(exp.experiment_id)
            logger.info(
                f"Experiment with ID {exp.experiment_id} and Name {exp.name} has been deleted"
            )
        experiments = ProgressLogger.get_experiments()
        assert len(experiments) == 0

    @staticmethod
    def get_experiments() -> List[Experiment]:
        return [
            e
            for e in mlflow.search_experiments()
            if e.experiment_id != FileStore.DEFAULT_EXPERIMENT_ID
        ]



import dataclasses
from typing import Optional, List

from dataclasses_json import DataClassJsonMixin

from helixcore.utilities.telemetry.telemetry_provider import TelemetryProvider
from helixcore.utilities.telemetry.telemetry_tracers import (
    TelemetryTracer,
)


@dataclasses.dataclass
class TelemetryContext(DataClassJsonMixin):
    telemetry_provider: str
    """ Provider for the telemetry context.  There must be corresponding telemetry implementation class registered for the provider """

    service_name: str
    """ Service name for the telemetry context """

    environment: str
    """ Environment for the telemetry context """

    trace_id: Optional[str] = None
    """ Trace ID for the telemetry context """

    span_id: Optional[str] = None
    """ Span ID for the telemetry context """

    trace_all_calls: Optional[List[TelemetryTracer]] = None
    """ Whether to Trace certain calls like aiohttp, pymysql, etc """

    tracer_endpoint: Optional[str] = None
    """ Tracer endpoint for the telemetry context """

    metrics_endpoint: Optional[str] = None
    """ Metrics endpoint for the telemetry context """

    @staticmethod
    def get_null_context() -> "TelemetryContext":
        """
        Get a null telemetry context

        :return: a null telemetry context
        """
        return TelemetryContext(
            telemetry_provider=TelemetryProvider.NULL,
            trace_id=None,
            span_id=None,
            service_name="",
            environment="",
        )

    def copy(self) -> "TelemetryContext":
        """
        Create a copy of the telemetry context

        :return: a copy of the telemetry context
        """
        return TelemetryContext(
            telemetry_provider=self.telemetry_provider,
            trace_id=self.trace_id,
            span_id=self.span_id,
            service_name=self.service_name,
            environment=self.environment,
            trace_all_calls=self.trace_all_calls,
        )

    def create_child_context(
        self, *, trace_id: Optional[str], span_id: Optional[str]
    ) -> "TelemetryContext":
        """
        Create a child telemetry context

        :param trace_id: trace ID for the child context
        :param span_id: span ID for the child context
        :return: a child telemetry context
        """
        telemetry_context = self.copy()
        telemetry_context.trace_id = trace_id
        telemetry_context.span_id = span_id
        return telemetry_context



import dataclasses
from typing import Optional

from dataclasses_json import DataClassJsonMixin


@dataclasses.dataclass
class GetBatchError(DataClassJsonMixin):
    url: str
    status_code: int
    error_text: str
    request_id: Optional[str]
